
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mindaffectBCI.utopiaclient import *

class UtopiaController:
    '''controller class to manage the interaction with the Mindaffect decoder,
    setting up the connection, sending and recieving messages, and firing message
    event handlers'''
    def __init__(self, clientid=None):
        self.client = UtopiaClient(clientid)
        self.msgs = []
        # callback list for new messages
        self.messageHandlers = []
        # call back list for new predictions
        self.lastPrediction = None
        self.predictionHandlers = []
        # selection stuff
        self.selectionHandlers = []
        self.selectionThreshold = .1
        # signal quality stuff
        self.lastSignalQuality = None
        self.signalQualityHandlers = []
        # our set message subscriptions
        self.subscriptions = "PSNMEQ"

    def addMessageHandler(self, cb):
        self.messageHandlers.append(cb)
    def addPredictionHandler(self, cb):
        self.predictionHandlers.append(cb)
    def addSelectionHandler(self, cb):
        self.selectionHandlers.append(cb)
    def addSignalQualityHandler(self, cb):
        self.signalQualityHandlers.append(cb)
    
    def setTimeStampClock(self,tsclock):
        return self.client.setTimeStampClock(tsclock)
    def getTimeStamp(self):
        '''get a (relative) wall-time stamp *in milliseconds*'''
        return self.client.getTimeStamp()
        
    def autoconnect(self, host=None, port=8400, timeout_ms=5000, 
                    localhostifhostnotfound=True, queryifhostnotfound=True, scanifhostnotfound=False):
        try:
            self.client.autoconnect(host, port, timeout_ms=timeout_ms, localhostifhostnotfound=localhostifhostnotfound, queryifhostnotfound=queryifhostnotfound, scanifhostnotfound=scanifhostnotfound)
        except socket.error as ex:
            print("Socket error connecting: ")
            print(ex)
            pass

        if not self.client.isConnected:
            print("Warning:: couldnt connect to a utopia hub....")
        else:
            # set default subscriptions
            self.subscribe()        

    def isConnected(self):  return self.client.isConnected
    def gethostport(self):  return self.client.gethostport()
                
    def sendStimulusEvent(self, stimulusState, timestamp=None, 
                          targetState=None, objIDs=None):
        """Send a message to the Utopia-HUB informing of the current stimulus state"""
        stimEvent = self.mkStimulusEvent(stimulusState, timestamp, targetState, objIDs)
        if self.client: self.client.sendMessage(stimEvent)
        # erp injection for debugging with fakedata
        if targetState in (0, 1): # TODO []: inject to the same host as the utopia connection
            injectERP(targetState)# , self.gethostport())
        return stimEvent
        
    def mkStimulusEvent(self, stimulusState, timestamp=None, 
                        targetState=None, objIDs=None):
        """make a valid stimulus event for the given stimulus state"""
        if timestamp is None:
            timestamp = self.getTimeStamp()
        if objIDs is None:
            objIDs = list(range(1, len(stimulusState)+1))
        elif len(objIDs) != len(stimulusState):
            raise ValueError("ARGH! objIDs and stimulusState not same length!")
    
        # insert extra 0 object ID if targetState given
        if not targetState is None and targetState >= 0:
            # N.B. copy longer version to new variable, rather than modify in-place with append
            objIDs = objIDs+[0] 
            stimulusState = stimulusState+[targetState]
        
        return StimulusEvent(timestamp, objIDs, stimulusState)

    def modeChange(self, newmode):
        if self.client:
            self.client.sendMessage(
                ModeChange(self.getTimeStamp(), newmode))

    def subscribe(self, msgs=None):
        # subscribe to PREDICTEDTARGETPROB, MODECHANGE, SELECTION and NEWTARGET, SIGNALQUALITY messages only
        if msgs:
            self.subscriptions = msgs
        if self.client:
            print("NewSubscriptions: {}".format(self.subscriptions))
            self.client.sendMessage(
                Subscribe(self.getTimeStamp(), self.subscriptions))
    def addSubscription(self, msgs):
        # N.B. we allow multiple subscribe to same message type so can remove without worrying about breaking
        # for another user
        for m in msgs:
            self.subscriptions += m
        self.subscribe()
    def removeSubscription(self, msgs):
        # remove msgs from the list of subscribed messages
        for m in msgs:
            self.subscriptions = self.subscriptions.replace(m, "", 1) # remove single subscription
        self.subscribe()


    def log(self, msg):
        if self.client:
            self.client.sendMessage(Log(self.getTimeStamp(), msg))

    def newTarget(self):
        if self.client:
            self.client.sendMessage(NewTarget(self.getTimeStamp()))

    def selection(self, objID):
        if self.client:
            self.client.sendMessage(Selection(self.getTimeStamp(), objID))
        for h in self.selectionHandlers:
            h(objID)         # do selection callbacks
            
    def getNewMessages(self, timeout_ms=0):
        '''get new messages from the utopia-hub, and store the list of new'''
        if not self.client: return None
        # get any messages with predictions
        self.msgs = self.client.getNewMessages(timeout_ms) if self.client else []
        # process these messages as needed & call-callbacks
        if len(self.msgs) > 0:
            for h in self.messageHandlers:
                h(self.msgs)
            newPrediction = None
            for m in self.msgs:
   
                if m.msgID == PredictedTargetProb.msgID:
                    # process new prediction callbacks
                    for h in self.predictionHandlers:
                        h(m)
                    # record as last prediction
                    #if m.Yest < 0:
                    #    m.Perr = 1
                    self.lastPrediction = m
                    print("Pred:{}".format(m))
                        
                elif m.msgID == Selection.msgID:
                    # process selection callbacks
                    for h in self.selectionHandlers:
                        h(m.objID)

                elif m.msgID == SignalQuality.msgID:
                    self.lastSignalQuality = m.signalQuality
                    # process selection callbacks
                    for h in self.signalQualityHandlers:
                        h(self.lastSignalQuality)

        return self.msgs

    def getLastPrediction(self):
        '''check for new predictions from the utopia-decoder'''
        # Q: should we do this here? or just return the lastPrediction?
        self.getNewMessages()
        # always return the last prediction, even if no new ones
        return self.lastPrediction
    
    def clearLastPrediction(self):
        '''clear the last predicted target'''
        self.lastPrediction = None
            
    def getLastSelection(self):
        """check if any object prediction is high enough for it to be selected"""
        self.getNewMessages()
        if self.lastPrediction is not None:
            if self.lastPrediction.Perr < self.selectionThreshold: # good enough to select?
                return (self.lastPrediction.Yest, True)

            else: # return predictedObjID but not-selected
                return (self.lastPrediction.Yest, False)
        return (None, False)

    def getLastSignalQuality(self):
        self.getNewMessages()
        return self.lastSignalQuality

def injectERP(amp, host="localhost", port=8300):
    """Inject an erp into a simulated data-stream, sliently ignore if failed, e.g. because not simulated"""
    import socket
    try:
        socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0).sendto(bytes([amp]), (host, port))
    except: # sliently igore any errors
        pass


def newMessageHandler(msgs):
    for m in msgs:
        print(m)

if __name__ == "__main__":
    # simple message logging testcase
    uc = UtopiaController()
    uc.autoconnect()
    # add logging incomming message handler
    uc.addMessageHandler(newMessageHandler)
    while uc.isConnected():
        uc.getNewMessages(1000)
        #msgs = uc.getNewMessages(1000)
        # print the recieved messages
        #for m in msgs:
        #    print(m)



#!/usr/bin/env python3
import numpy as np
from mindaffectBCI.decoder.UtopiaDataInterface import UtopiaDataInterface, stim2eventfilt, butterfilt_and_downsample
from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics
from mindaffectBCI.decoder.multipleCCA import multipleCCA
from mindaffectBCI.decoder.stim2event import stim2event
from mindaffectBCI.decoder.devent2stimsequence import upsample_stimseq

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

def ccaViewer(ui: UtopiaDataInterface, timeout_ms: float=np.inf, tau_ms: float=500,
              offset_ms=(-15, 0), evtlabs=None, ch_names=None, nstimulus_events: int=600, rank=3, reg=.02, center=True):
    ''' simple sig-viewer using the ring-buffer for testing '''


    if evtlabs is None:
        evtlabs = ('re', 'fe')

    ui.update()

    # compute the size of the erp slice
    irf_range_ms = (offset_ms[0], tau_ms+offset_ms[1])
    irflen_ms = irf_range_ms[1]-irf_range_ms[0]
    irflen_samp = int(np.ceil(irflen_ms * ui.fs / 1000))
    irf_times = np.linspace(irf_range_ms[0], irf_range_ms[1], irflen_samp)

    # initialize the arrays to hold the stimulus responses, and their label info, and a current cursor
    irf = np.zeros((nstimulus_events, irflen_samp, ui.data_ringbuffer.shape[-1] - 1)) # (nERP,irflen,d)
    irf_lab = np.zeros((nstimulus_events, len(evtlabs)), dtype=int)  # (nErp,nstim)
    nY = np.zeros(len(evtlabs),dtype=int)

    # store for the summary statistics
    Cxx = None
    Cxy = None
    Cyy = None
    
    # initialize the plot window
    fig = plt.figure(1)
    fig.clear()

    # main spec, inc titles
    outer_grid = fig.add_gridspec(nrows=2, ncols=2, height_ratios=[1,6])

    # right-sub-spec for the ERPs
    plt.figtext(.25,.9,'ERPs',ha='center')
    # get grid-spec to layout the plots, 1 for spatial, 1 for temporal
    gs = outer_grid[-1,0].subgridspec(nrows=len(evtlabs), ncols=1)
    erp_ax = [None for j in range(len(evtlabs))]
    erp_lines = [None for j in range(len(evtlabs))]
    erp_ax[-1] = fig.add_subplot(gs[-1,0])
    for ei,lab in enumerate(evtlabs):
        if ei==len(evtlabs)-1:
            ax = erp_ax[ei]
            ax.set_xlabel("Time (ms)")
            ax.set_ylabel("Space")
        else:
            ax = fig.add_subplot(gs[ei,0], sharex=erp_ax[-1], sharey=erp_ax[-1])
            ax.tick_params(labelbottom=False)
        erp_lines[ei] = ax.imshow(irf[ei,:,:].T,aspect='auto',origin='lower',extent=(irf_times[0],irf_times[-1],0,irf.shape[-1]))
        ax.set_title(lab)
        erp_ax[ei] = ax

    # left-sub-spec for the decomposition
    plt.figtext(.75,.9,'CCA Decomposition',ha='center')
    # get grid-spec to layout the plots, 1 for spatial, 1 for temporal
    gs = outer_grid[-1,1].subgridspec(nrows=rank, ncols=2, width_ratios=[3,1])
    spatial_ax = [ None for i in range(rank)]
    spatial_lines = [ None for i in range(rank)] 
    temporal_ax = [None for j in range(rank)]
    temporal_lines = [[None for i in range(len(evtlabs))] for j in range(rank)]

    # make the bottom 2 axs already so can share their limits.
    temporal_ax[-1] = fig.add_subplot(gs[-1, 0])
    spatial_ax[-1] = fig.add_subplot(gs[-1, 1])
    for ri in range(rank):
        # single temporal plot for all ranks
        if ri==rank-1: # tick-plot
            ax = temporal_ax[-1]
        else:
            ax = fig.add_subplot(gs[ri, 0], sharex=temporal_ax[-1], sharey=temporal_ax[-1])
            ax.tick_params(labelbottom=False)
        # setup the lines
        for ei, lab in enumerate(evtlabs):
            # component range
            temporal_lines[ri][ei] = ax.plot(irf_times, np.zeros(irf_times.shape), label=lab)[0]
            # len(evtlabs))))
            #temporal_lines[ri][ei].set_label(evtlabs[ei])
        ax.set_ylabel("Comp #{}".format(ri))    
        ax.grid(True)
        #ax.autoscale(axis='y')
        if ri==rank-1:
            ax.set_xlabel("Time (ms)") 
            ax.legend()
        temporal_ax[ri] = ax

        # spatial-plot
        if ri==rank-1: # tick-plot
            ax = spatial_ax[ri]
            ax.set_xlabel("Space")
        else:
            ax = fig.add_subplot(gs[ri, 1], sharex=spatial_ax[-1], sharey=spatial_ax[-1])
            ax.tick_params(labelbottom=False)

        ax.grid(True)
        spatial_ax[ri] = ax
        spatial_lines[ri] = ax.plot(np.zeros((irf.shape[-1])), color=(0, 0, 0))

    # add a reset ERP button
    def reset(event):
        print('reset called')
        nonlocal Cxx, Cxy, Cyy
        Cxx=None
        Cxy=None
        Cyy=None
        nY.fill(0)

    from matplotlib.widgets import Button
    butax = fig.add_axes([.05,.85,.1,.1])
    breset = Button(butax,'Reset')
    breset.on_clicked(reset)

    # tidy up the axes locations
    plt.tight_layout()
    fig.show()
    
    # start the render loop
    t0 = ui.getTimeStamp()
    block_start_ts = t0
    oM = None  # last few stimulus states, for correct transformation of later events
    while ui.getTimeStamp() < t0+timeout_ms:

        # exit if figure is closed..
        if not plt.fignum_exists(1):
            quit()

        # re-draw the display
        fig.canvas.draw()
        fig.canvas.flush_events()
        #plt.pause(.001)

        # Update the records
        nmsgs, ndata, nstim = ui.update(timeout_ms=100)
        if nstim == 0:
            continue
        
        # last time for which we have full response for a stimulus
        valid_end_ts = min(ui.stimulus_timestamp + irflen_ms, ui.data_timestamp)

        # skip if no new data to process
        if block_start_ts + irflen_ms > valid_end_ts:
            continue 

        block_end_ts = valid_end_ts

        # extract and apply to this block
        print("Extract block: {}->{} = {}ms".format(block_start_ts, block_end_ts, block_end_ts-block_start_ts))
        data = ui.extract_data_segment(block_start_ts, block_end_ts)
        stimulus = ui.extract_stimulus_segment(block_start_ts, block_end_ts)
        block_start_ts = data[-irflen_samp,-1] # block_end_ts - irflen_ms

        # skip if no data/stimulus to process
        if data.size == 0 or stimulus.size == 0:
            continue

        # upsample stimulus to sample rate
        X_ts = data[:, -1]
        X = data[:, :-1]
        M_ts = stimulus[:, -1]
        M = stimulus[:, :-1]
        M, _ = upsample_stimseq(X_ts, M, M_ts)

        # TODO[]: limit to active outputs

        # and transform to irf trigger events
        Y = stim2event(M, evtypes=evtlabs, axis=-2, oM=oM)  # (nsamp,nstim,e)
        # record the last bit before the processed M for next time, i.e. 
        oM = M[:-irflen_samp, :]

        # strip to only the objID==0 output
        Y = Y[:, 0:1, :]  # (nsamp,1,nE)
        # count the numbers each type of event
        nY = nY + np.sum(Y,axis=(0,1))

        # incrementally update the summary statistics
        Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y, None, Cxx, Cxy, Cyy, tau=irflen_samp)

        # debug plot summary statistics.
        erp_lim = (np.min(Cxy.ravel()),np.max(Cxy.ravel()))
        for ei,lab in enumerate(evtlabs):
            erp_lines[ei].set_data(Cxy[0,ei,:,:].T)
            erp_lines[ei].set_clim(vmin=erp_lim[0],vmax=erp_lim[1])
            #erp_ax[ei].imshow(Cxy[0,ei,:,:].T/nY[ei],aspect='auto',extent=(irf_times[0],irf_times[-1],0,Cxy.shape[-1]))
            erp_ax[ei].set_title("{} (n={})".format(lab,nY[ei]))

        # update the cca decomposition
        J, W, R = multipleCCA(Cxx, Cxy, Cyy, reg=reg, rank=rank)
        # strip the model dimension
        J = J[0]
        W = W[0,...]
        R = R[0,...]

        R_lim = (-np.max(np.abs(R.ravel())) * 2, np.max(np.abs(R.ravel())) * 2)
        R_lim = [ d if not np.isnan(d) else 0 for d in R_lim ] # guard

        W_lim = (-np.max(np.abs(W.ravel()))*2, np.max(np.abs(W.ravel()))*2)
        W_lim = [ d if not np.isnan(d) else 0 for d in W_lim ] # guard

        # Update the plots for each event type
        for ri in range(rank):
            sign = np.sign(W[ri,np.argmax(np.abs(W[ri,:]))]) # normalize directions

            # plot the temporal responses
            for ei in range(len(evtlabs)):
                temporal_lines[ri][ei].set_ydata(R[ri,ei,:]*sign)
            temporal_ax[ri].set_ylim( R_lim )

            # plot the spatial patterns
            # TODO[]: as a topographic plot
            spatial_lines[ri][-1].set_ydata(W[ri,:]*sign)
            spatial_ax[ri].set_ylim(W_lim)

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, help='address (IP) of the utopia-hub', default=None)
    parser.add_argument('--evtlabs', type=str, help='comma separated list of stimulus even types to use', default='re,fe')
    parser.add_argument('--rank', type=str, help='rank of decomposition to use', default=3)
    parser.add_argument('--ch_names', type=str, help='list of channel names, or capfile', default=None)
    args = parser.parse_args()
    hostname = args.host
    evtlabs = args.evtlabs.split(',')
    rank = args.rank
    print('evtlabs={}'.format(evtlabs))
    ch_names = args.ch_names.split(',') if args.ch_names is not None else None

    data_preprocessor = None
    data_preprocessor = butterfilt_and_downsample(order=4, stopband=((45,65), (0, 4), (25, -1)), fs_out=100)
    ui=UtopiaDataInterface(data_preprocessor=data_preprocessor, send_signalquality=False)
    ui.connect(hostname)

    try:
        matplotlib.use('Qt4Cairo')
    except:
        pass

    ccaViewer(ui, evtlabs=evtlabs, ch_names=ch_names, rank=rank)



#!/usr/bin/env python3
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import random
from .stimseq import StimSeq
from .utopiaController import UtopiaController, TimeStampClock
import os
scriptpath=os.path.dirname(os.path.realpath(__file__))
default_stimFile=os.path.join(scriptpath,'mgold_61_6521_psk_60hz.txt')
objIDs=list(range(1,10))
isi=1/60
MAXOBJID=64

class FSM:
    ''' simple finite state machine, using a generator-like pattern'''
    def next(self,t):
        '''update the current state, return the new state, raise StopIteration exception when done'''
        return t
    def get(self):
        '''get the current state, for this set of active objects'''
        return (None,-1,None,False)# BODGE: stimState,tgtState,objIDs,sendEvent

class GSM(FSM):
    '''Generalized state machine with stack of states'''
    def __init__(self): self.stack=[]
    def clear(self):    self.stack=[]
    def push(self,s):   self.stack.append(s); return self.stack
    def pop(self):      return self.stack.pop()
    def next(self,t):
        '''get the next stimulus state to shown'''
        while self.stack :
            try : 
                self.stack[-1].next(t)
                return
            except StopIteration :
                # end of this fsm, unwind the fsm stack
                self.pop()
                # for pretty printing
                print()
        raise StopIteration()
    def get(self):
        if self.stack :
            return self.stack[-1].get()
        else :
            return None

class WaitFor(FSM):
    '''wait for given number of frames to pass'''
    def __init__(self,numframes):
        self.numframes=numframes
        self.nframe=0
        print("waitFor: %g"%(self.numframes))
    def next(self,t):
        self.nframe=self.nframe+1
        if self.nframe>self.numframes :
            raise StopIteration()
    def get(self):
        return (None,-1,None,False) # BODGE: stimState,tgtState,objIDs,sendEvent
    
class Flicker(FSM):
    ''' do a normal flicker sequence'''
    def __init__(self,stimSeq=None,
                 numframes=4*isi, # number for frames to flicker for
                 tgtidx=-1, # target output for this flicker
                 sendEvents=True, # send stimulus events
                 framesperbit=1): # number of video-frames per codebook bit 
        self.stimSeq=stimSeq
        self.numframes=numframes
        self.nframe=0
        self.tgtidx=tgtidx
        self.tgtstate=-1
        self.sendEvents=sendEvents
        self.framesperbit=framesperbit if framesperbit is not None else 1

        # ensure right length
        self.ss=None
        #print("stimSeq:")
        #for i in range(len(self.objIDs)):
        #    print(["objID %d = %s"%(i,"".join([ '*' if self.stimSeq[t][i]==1 else '.' for t in range(len(self.stimSeq))]))])
        print('flicker: %d frames, tgt %d'%(self.numframes,tgtidx if tgtidx >=0 else -1))
        
    def next(self,t):
        self.nframe=self.nframe+1
        if self.nframe>self.numframes:
            raise StopIteration()

    def update_ss(self):
        # extract the current frames stimulus state, loop if past end
        if self.nframe >= self.numframes:
            # final frame is blank screen
            self.ss[:] = [0]*len(self.stimSeq[0])
            self.tgtState = -1
        else:            
            self.ss       = self.stimSeq[self.nframe//self.framesperbit % len(self.stimSeq)]
            # extract the current target state, for these objects
            self.tgtstate = self.ss[self.tgtidx] if self.tgtidx>=0 else -1
        
    def get(self):
        # update the curent stimulus state info
        self.update_ss()        
        # default to the stored set active objIDs
        #if objIDs is None : objIDs=self.objIDs
        # get the state info for the given set of objectIDs
        #ss =[self.ss[i] for i in objIDs-1]
        return (self.ss,self.tgtstate,None,self.sendEvents)

    
class FlickerWithSelection(Flicker):
    ''' do a normal flicker sequence, with early stopping selection'''
    def __init__(self,
                 stimSeq=None,
                 numframes=4*isi,
                 tgtidx=-1,
                 utopiaController=None,
                 framesperbit=1,
                 sendEvents=True):
        super().__init__(stimSeq,numframes,tgtidx,sendEvents,framesperbit)
        self.utopiaController = utopiaController
        if self.utopiaController is None : raise ValueError("must have utopiaController")
        # ensure old predictions are gone..
        self.utopiaController.getNewMessages() # ensure all old predictions processed
        self.utopiaController.clearLastPrediction() # reset to no waiting predictions
        print(' with selection')
        
    def next(self,t):
        super().next(t)
        # check for selection and stop if found
        objId,selected=self.utopiaController.getLastSelection()
        if selected :
            if self.sendEvents :
                # send event to say selection has occured
                self.utopiaController.selection(objId)
            raise StopIteration()

def mkBlinkingSequence(numframes,tgtidx,tgtState=1):
    blinkSeq=[[0 for i in range(MAXOBJID)] for i in range(numframes)]
    for i in range(int(numframes/2)) : blinkSeq[i][tgtidx]=tgtState # tgt on
    return blinkSeq

class HighlightObject(Flicker):
    '''Highlight a single object for a number of frames'''
    def __init__(self,numframes=isi*2,tgtidx=-1,tgtState=2,
                 sendEvents=False,numblinkframes=int(.5/isi)):
        #self.objIDs=objIDs if hasattr(objIDs, "__len__") else list(range(1,objIDs+1))
        if numblinkframes>0 and tgtidx>=0 : # blinking cue
            stimSeq = mkBlinkingSequence(int(numblinkframes),tgtidx,tgtState)
        else :
            stimSeq = [[0]*MAXOBJID]
            if tgtidx>=0:
                stimSeq[0][tgtidx]=tgtState
        super().__init__(stimSeq,numframes,tgtidx,sendEvents=sendEvents)
        print('highlight: tgtidx=%d nframes=%d'%(tgtidx if tgtidx>=0 else -1,numframes))

class SingleTrial(FSM):
    ''' do a complete single trial with: cue->wait->flicker->feedback '''
    def __init__(self,stimSeq,tgtidx,
                 utopiaController,stimulusStateStack,
                 numframes=None,framesperbit=1,
                 selectionThreshold=None,
                 duration=4,cueduration=1,feedbackduration=1,waitduration=1,
                 cueframes=None,feedbackframes=None,waitframes=None):
        self.tgtidx=tgtidx
        self.stimSeq=stimSeq
        self.utopiaController = utopiaController
        self.stimulusStateStack=stimulusStateStack
        self.numframes=numframes if numframes else duration/isi
        self.cueframes=cueframes if cueframes else cueduration/isi
        self.framesperbit=framesperbit if  framesperbit is not None else 1
        self.feedbackframes=feedbackframes if feedbackframes else feedbackduration/isi
        self.waitframes=waitframes if waitframes else waitduration/isi
        self.selectionThreshold=selectionThreshold
        self.stage=0
        self.stagestart = self.utopiaController.getTimeStamp()
        print("tgtidx=%d"%(self.tgtidx if self.tgtidx>=0 else -1))
        
    def next(self,t):
        last_stage_dur = self.utopiaController.getTimeStamp()-self.stagestart
        if self.stage==0 : # trial-start + cue
            # tell decoder to start trial
            self.utopiaController.newTarget()
            # tell decoder to clear predictions if needed
            if self.selectionThreshold is not None:
                self.utopiaController.selectionThreshold=self.selectionThreshold
                self.utopiaController.clearLastPrediction() # reset to no waiting predictions

            if self.tgtidx>=0 : # blinking: only if target is set
                self.stimulusStateStack.push(
                    HighlightObject(self.cueframes,
                                    self.tgtidx,
                                    sendEvents=False))
            else : # skip cue+wait
                self.stage=1
                
        elif self.stage==1 : # wait
            print('1.wait')
            self.stimulusStateStack.push(
                HighlightObject(self.waitframes,-1,
                                sendEvents=False))
            
        elif self.stage==2 : # stim
            print('2.stim, tgt:%d'%(self.tgtidx if self.tgtidx>=0 else -1))
            #objIDs=8,stimSeq=None,numframes=None,tgtidx=None,duration=4
            if self.selectionThreshold is not None: # early stop if thres set
                self.stimulusStateStack.push(
                    FlickerWithSelection(self.stimSeq,
                                         self.numframes,
                                         self.tgtidx,
                                         self.utopiaController,
                                         framesperbit=self.framesperbit,
                                         sendEvents=True))
            else: # no selection based stopping
                self.stimulusStateStack.push(
                    Flicker(self.stimSeq,
                            self.numframes,
                            self.tgtidx,
                            framesperbit=self.framesperbit,
                            sendEvents=True))
                
        elif self.stage==3 : # wait/feedback
            if self.selectionThreshold is None:
                print('3.wait')
                self.stimulusStateStack.push(
                    HighlightObject(self.waitframes,-1))
            else :
                print('3.feedback') # solid on 
                predObjId,selected=self.utopiaController.getLastSelection()
                print('%dms pred:%d sel:%d  %c'%(int(last_stage_dur),
                                                 predObjId-1 if predObjId else -1,selected,
                                                 "*" if predObjId and self.tgtidx==predObjId-1 else "x"))
                if selected :
                    #tgtidx = self.objIDs.index(predObjId) if predObjId in self.objIDs else -1
                    tgtidx = predObjId-1
                    self.stimulusStateStack.push(
                        HighlightObject(self.feedbackframes,
                                        tgtidx,
                                        tgtState=3,
                                        sendEvents=False,
                                        numblinkframes=0))
                    
        else :
            raise StopIteration
        self.stagestart = self.utopiaController.getTimeStamp()
        self.stage=self.stage+1
        
class CalibrationPhase(FSM):
    ''' do a complete calibration phase with nTrials x CalibrationTrial '''
    def __init__(self,objIDs=8,stimSeq=None,nTrials=10,
                 utopiaController=None,stimulusStateStack=None,
                 *args,**kwargs):
        self.objIDs=objIDs if hasattr(objIDs, "__len__") else list(range(1,objIDs+1)) 
        self.stimSeq=stimSeq
        self.nTrials=nTrials
        self.utopiaController=utopiaController
        self.isRunning=False
        self.args=args
        self.kwargs=kwargs
        if self.utopiaController is None : raise ValueError
        self.stimulusStateStack=stimulusStateStack
        if self.stimulusStateStack is None : raise ValueError
        self.trli=0
        self.tgtidx = -1

    def next(self,t):
        if not self.isRunning :
            # tell decoder to start cal
            self.utopiaController.modeChange("Calibration.supervised")
            self.isRunning=True
        if self.trli<self.nTrials:
            # TODO []: should choose from set active objIDs?
            tgtidx = random.randint(0,len(self.objIDs)-1)
            self.tgtidx = tgtidx if not tgtidx == self.tgtidx else (tgtidx+1)%len(self.objIDs)
            print("Start Cal: %d/%d tgtidx=%d"%(self.trli,self.nTrials,self.tgtidx))
            self.stimulusStateStack.push(
                SingleTrial(self.stimSeq,
                            self.tgtidx,
                            self.utopiaController,
                            self.stimulusStateStack,
                            *self.args,**self.kwargs)) # pass through other arguments?
        else:
            self.utopiaController.modeChange("idle")
            raise StopIteration()
        self.trli=self.trli+1
            
class PredictionPhase(FSM):
    '''do complete prediction phase with nTrials x trials with early-stopping feedback'''
    def __init__(self,objIDs,stimSeq=None,nTrials=10,
                 utopiaController=None,stimulusStateStack=None,
                 selectionThreshold=.1,cuedprediction=False,*args,**kwargs):
        self.objIDs=objIDs if hasattr(objIDs, "__len__") else list(range(1,objIDs+1)) 
        self.stimSeq=stimSeq
        self.nTrials=nTrials
        self.selectionThreshold=selectionThreshold
        self.cuedprediction=cuedprediction
        self.args=args
        self.kwargs=kwargs
        self.tgti=0
        self.tgtidx = -1
        self.isRunning=False
        self.utopiaController=utopiaController
        if self.utopiaController is None : raise ValueError
        self.stimulusStateStack=stimulusStateStack
        if self.stimulusStateStack is None : raise ValueError
        # tell decoder to start cal

    def next(self,t):
        if not self.isRunning :
            self.utopiaController.modeChange("Prediction.static")
            self.isRunning=True
        if self.tgti<self.nTrials:
            if self.cuedprediction :
                tgtidx = random.randint(0,len(self.objIDs)-1)
                self.tgtidx = tgtidx if not tgtidx == self.tgtidx else (tgtidx+1)%len(self.objIDs)
            else:
                self.tgtidx = -1
            print("Start Pred: %d/%d"%(self.tgti,self.nTrials))
            self.stimulusStateStack.push(
                SingleTrial(self.stimSeq,
                            self.tgtidx,
                            self.utopiaController,
                            self.stimulusStateStack,
                            selectionThreshold=self.selectionThreshold,
                            *self.args,**self.kwargs))
        else:
            self.utopiaController.modeChange("idle")
            raise StopIteration()
        self.tgti=self.tgti+1
        
class Experiment(FSM):
    '''do a complete experiment, with calibration -> prediction'''
    def __init__(self,objIDs,stimSeq=None,nCal=10,nPred=20,
                 selectionThreshold=.1,cuedprediction=False,
                 utopiaController=None,stimulusStateStack=None,
                 duration=4,calduration=4,predduration=10,
                 *args,**kwargs):
        self.objIDs=objIDs if hasattr(objIDs, "__len__") else list(range(1,objIDs+1)) 
        self.stimSeq=stimSeq
        self.nCal=nCal
        self.nPred=nPred
        self.calduration=calduration
        self.predduration=predduration
        self.selectionThreshold=selectionThreshold
        self.cuedprediction=cuedprediction
        self.utopiaController=utopiaController
        if self.utopiaController is None : raise ValueError
        self.stimulusStateStack=stimulusStateStack
        if self.stimulusStateStack is None : raise ValueError
        self.args=args
        self.kwargs=kwargs
        self.stage=0

    def next(self,t):
        if self.stage==0:
            self.stimulusStateStack.push(WaitFor(2/isi))
        
        elif self.stage==1:
            self.stimulusStateStack.push(
                CalibrationPhase(self.objIDs,self.stimSeq,self.nCal,
                                 self.utopiaController,
                                 self.stimulusStateStack,
                                 *self.args,duration=self.calduration,**self.kwargs))
        
        elif self.stage==2:
            self.stimulusStateStack.push(WaitFor(15/isi))
        
        elif self.stage==3:
            self.stimulusStateStack.push(
                PredictionPhase(self.objIDs,self.stimSeq,self.nPred,
                                self.utopiaController,
                                self.stimulusStateStack,
                                self.selectionThreshold,
                                self.cuedprediction,
                                *self.args,duration=self.predduration,**self.kwargs))
        
        else:
            raise StopIteration()    
        self.stage=self.stage+1

# use global for the noisecode + utopiaController so can easily
# share over instantiations of the NoiseTag framework.
uc=None
        
class Noisetag:
    '''noisetag abstraction layer to handle *both* the sequencing of the stimulus
    flicker, *and* the communications with the Mindaffect decoder.  Clients can
    use this class to implement BCI control by:
     0) setting the flicker sequence to use (method: startFlicker, startFlickerWithSelection, startCalibration, startPrediction, startExpt)
     1) getting the current stimulus state (method: getStimulusState), and using that to draw the display
     2) telling Noisetag when *exactly* the stimulus update took place (method: sendStimulusState)
     3) getting the predictions/selections from noisetag and acting on them. (method: getLastPrediction() or getLastSelection())
     '''
    def __init__(self,stimFile=None,utopiaController=None,stimulusStateMachineStack=None,clientid:str=None):
        # global flicker stimulus sequence
        if stimFile is None:  stimFile = default_stimFile
        noisecode = StimSeq.fromFile(stimFile)
        # binarize the result
        noisecode.convertstimSeq2int()
        self.noisecode=noisecode

        # utopiaController
        self.utopiaController = utopiaController
    
        # stimulus state-machine stack
        # Individual stimulus-state-machines track progress in a single
        # stimulus state playback function.
        # Stack allows sequencing of sets of playback functions in loops
        if stimulusStateMachineStack is None :
            stimulusStateMachineStack=GSM()
        self.stimulusStateMachineStack=stimulusStateMachineStack
        self.laststate=(None,None,None,None)
        self.objIDs=None
        self.clientid=clientid

    def connect(self,host=None,port=-1,queryifhostnotfound=True,timeout_ms=5000):
        if self.utopiaController is None :
            # use the global controller if none given
            global uc
            if uc is None :
                # auto-connect the global controller if none given
                uc = UtopiaController(clientid=self.clientid)
            self.utopiaController=uc
        if self.utopiaController.isConnected() :
            return True
        self.utopiaController.autoconnect(host,port,
                                          queryifhostnotfound=queryifhostnotfound,
                                          localhostifhostnotfound=True,
                                          timeout_ms=timeout_ms)
        return self.utopiaController.isConnected()
    
    def isConnected(self):
        if self.utopiaController :
            return self.utopiaController.isConnected()
        return False
    def gethostport(self):
        if self.utopiaController :
            return self.utopiaController.gethostport()
        return None
        
    # stimulus sequence methods via the stimulus state machine stack
    # returns if sequence is still running
    def updateStimulusState(self,t=None):
        self.stimulusStateMachineStack.next(t)

    def getStimulusState(self,objIDs=None):
        # update set active objects if a set is given
        if objIDs is not None : 
            self.setActiveObjIDs(objIDs)
        # get the complete stimulus state (for MAXOBIDS objects)
        stimState,tgtstate,objIDs,sendEvents = self.stimulusStateMachineStack.get()
        # subset to the active set, matching objIDs to allobjIDs
        # N.B. objID-1 to map from objID->stimStateIndex
        if stimState is not None :
            stimState = [ stimState[i-1] for i in self.objIDs ]
        self.laststate = (stimState,tgtstate,self.objIDs,sendEvents)
        return self.laststate
    
    def setActiveObjIDs(self,objIDs):
        '''update the set of active objects we send info to decoder about'''
        self.objIDs=objIDs
        return self.objIDs
    
    def setnumActiveObjIDs(self,nobj):
        '''update to say number active objects'''
        objIDs=list(range(1,nobj+1))
        return self.setActiveObjIDs(objIDs)
    
    # decoder interaction methods via. utopia controller
    def sendStimulusState(self,timestamp=None):
        stimState,targetState,objIDs,sendEvent=self.laststate
        # send info about the stimulus displayed
        if sendEvent and stimState is not None :
            #print((stimState,targetState))
            self.utopiaController.sendStimulusEvent(stimState,
                                                    timestamp,
                                                    targetState,
                                                    objIDs)

    def getNewMessages(self):
        if self.utopiaController:
            return self.utopiaController.msgs
        return []
    def getLastPrediction(self):
        if self.utopiaController :
            return self.utopiaController.getLastPrediction()
        return None
    def clearLastPrediction(self):
        if self.utopiaController:
            self.utopiaController.clearLastPrediction()
    def getLastSignalQuality(self):
        if self.utopiaController :
            return self.utopiaController.getLastSignalQuality()
        return None
    def clearLastSignalQuality(self):
        if self.utopiaController:
            self.utopiaController.clearLastSignalQuality()
    def getLastSelection(self):
        if self.utopiaController :
            return self.utopiaController.getLastSelection()
        return None
    def clearLastSelection(self):
        if self.utopiaController:
            self.utopiaController.clearLastSelection()
    def addMessageHandler(self,cb):
        if self.utopiaController :
            self.utopiaController.addMessageHandler(cb)
    def addPredictionHandler(self,cb):
        if self.utopiaController :
            self.utopiaController.addPredictionHandler(cb)
    def addSelectionHandler(self,cb):
        if self.utopiaController :
            self.utopiaController.addSelectionHandler(cb)
    
    def setTimeStampClock(self, tsclock):
        self.utopiaController.setTimeStampClock(tsclock)
    def getTimeStamp(self):
        return self.utopiaController.getTimeStamp() if self.utopiaController is not None else -1

    def log(self,msg):
        if self.utopiaController:
            self.utopiaController.log(msg)
    def modeChange(self,newmode):
        '''manually change the decoder mode'''
        if self.utopiaController:
            self.utopiaController.modeChange(newmode)
    def subscribe(self,msgs):
        if self.utopiaController:
            self.utopiaController.subscribe(msgs)
    def addSubscription(self,msgs):
        if self.utopiaController:
            self.utopiaController.addSubscription(msgs)
    def removeSubscription(self,msgs):
        if self.utopiaController:
            self.utopiaController.removeSubscription(msgs)

    # methods to define what (meta) stimulus sequence we will play
    def startExpt(self,nCal=1,nPred=20,selectionThreshold=.1,
                  cuedprediction=False,
                  *args,**kwargs):
        if  self.stimulusStateMachineStack.stack :
            print("Warning: replacing running sequence?")
            self.stimulusStateMachineStack.clear()
        self.stimulusStateMachineStack.push(
            Experiment(self.objIDs,self.noisecode.stimSeq,
                       nCal,nPred,
                       selectionThreshold,cuedprediction,
                       self.utopiaController,
                       self.stimulusStateMachineStack,
                       *args,**kwargs))
        
    def startCalibration(self,nTrials=10,stimSeq=None,
                         *args,**kwargs):
        if  self.stimulusStateMachineStack.stack :
            print("Warning: replacing running sequence?")
            self.stimulusStateMachineStack.clear()
        
        self.stimulusStateMachineStack.push(
            CalibrationPhase(self.objIDs,
                             self.noisecode.stimSeq,
                             nTrials,
                             self.utopiaController,
                             self.stimulusStateMachineStack,
                             *args,**kwargs))
        
    def startPrediction(self,nTrials=10,stimSeq=None,
                        *args,**kwargs):
        if  self.stimulusStateMachineStack.stack :
            print("Warning: replacing running sequence?")
            self.stimulusStateMachineStack.clear()
        self.stimulusStateMachineStack.push(
            PredictionPhase(self.objIDs,
                            self.noisecode.stimSeq,
                            nTrials,
                            self.utopiaController,
                            self.stimulusStateMachineStack,
                            *args,**kwargs))
        
    def startSingleTrial(self,numframes=100,tgtidx=-1,*args,**kwargs):
        if  self.stimulusStateMachineStack.stack :
            print("Warning: replacing running sequence?")
            self.stimulusStateMachineStack.clear()
        self.stimulusStateMachineStack.push(
            SingleTrial(self.noisecode.stimSeq,
                        tgtidx,
                        self.utopiaController,
                        self.stimulusStateMachineStack,
                        numframes,
                        *args,**kwargs))

    def startFlicker(self,numframes=100,tgtidx=-1,*args,**kwargs):
        if  self.stimulusStateMachineStack.stack :
            print("Warning: replacing running sequence?")
            self.stimulusStateMachineStack.clear()
        self.stimulusStateMachineStack.push(
            Flicker(self.noisecode.stimSeq,
                    numframes,
                    tgtidx,
                    *args,**kwargs))
        
    def startFlickerWithSelection(self,numframes=100,
                                  tgtidx=-1,*args,**kwargs):
        if  self.stimulusStateMachineStack.stack :
            print("Warning: replacing running sequence?")
            self.stimulusStateMachineStack.clear()
        self.stimulusStateMachineStack.push(
            FlickerWithSelection(self.noisecode.stimSeq,
                                 numframes,
                                 tgtidx,
                                 self.utopiaController,
                                 *args,**kwargs))


class sumstats:
    '''Utility class to record summary stastics for, e.g. frame flip timing'''
    def __init__(self,bufsize:int=60*2):
        self.buf=[0]*(bufsize) # ring-buffer, 700 entries
        self.N=0
        self.sx=0
        self.mu=-1
        self.sx2=0
        self.sigma2=-1
        self.minx=0
        self.maxx=0

    def addpoint(self,x):
        self.buf[self.N%len(self.buf)]=x # ring-buffer
        self.N=self.N+1
        self.sx=self.sx+x
        self.sx2=self.sx2+x*x
        self.minx=x if x<self.minx else self.minx
        self.maxx=x if x>self.maxx else self.maxx

    def hist(self):
        try:
            import numpy
            buf = self.buf[:min(len(self.buf),self.N)]
            self.minx=min(buf)
            self.maxx=max(buf)
            bstart=self.minx + (self.maxx-self.minx)*0
            bw    =(self.maxx-self.minx)*(.6-0)
            bins=[bstart+bw*(x/15.0) for x in range(0,15)]
            [hist,bins]=numpy.histogram(buf,bins)
            pp= " ".join("%5.2f"%((bins[i]+bins[i+1])/2) for i in range(len(bins)-1))
            pp+="\n" + " ".join("%5d"%t for t in hist)
        except:
            pp=''
        return pp

    def update_statistics(self):
        import statistics
        buf = self.buf[:min(len(self.buf),self.N)]
        self.mu = statistics.mean(buf)
        self.med= statistics.median(buf)
        self.sigma=statistics.stdev(buf) if len(buf)>2 else -1
        self.min = min(buf)
        self.max = max(buf)

    def __str__(self):
        self.update_statistics()
        return "%f,%f (%f,[%f,%f])"%(self.mu,self.median,self.sigma,self.min,self.max)

def doFrame(t,stimState,tgtState=-1,objIDs=None,utopiaController=None):
    if tgtState>=0 :
        print("*" if tgtState>0 else ".",end='',flush=True)
    else:
        print('_',end='',flush=True)
    
if __name__ == "__main__":
    # make the noisetag object to manage the tagging selections
    ntexpt = Noisetag()
    ntexpt.connect()
    # set the subset of active objects being displayed
    ntexpt.setnumActiveObjIDs(10)
    # tell it to play a full experiment sequence
    ntexpt.startExpt(framesperbit=4)
    # mainloop
    nframe=0
    while True:
        try :
            # update the stimulus state w.r.t. current time
            ntexpt.updateStimulusState(nframe)
            # get the stimulus state info we whould display
            ss,ts,objIDs,sendEvent=ntexpt.getStimulusState()
            # update the display drawing
            doFrame(nframe,ss,ts,objIDs)            
            # send info about what we did to the decoder
            ntexpt.sendStimulusState()
            # simulate waiting for the flip
            time.sleep(isi)
        except StopIteration :
            # this event is raised when the stimulus sequence is finished
            break
        nframe=nframe+1



import subprocess
import os

def run(label=''):
    pydir = os.path.dirname(os.path.abspath(__file__)) # mindaffectBCI/decoder/startUtopiaHub.py
    bindir = os.path.join(pydir,'..','..','bin') # ../../bin/

    # make the logs directory if not already there
    try: 
        os.mkdir(os.path.join(bindir,'..','logs'))
    except:
        pass

    # command to run the java hub
    cmd = ("java","-jar","UtopiaServer.jar")
    # args to pass to the java hub
    if label is not None:
        logfile = "../logs/mindaffectBCI_{}.txt".format(label)
    else:
        logfile = "../logs/mindaffectBCI.txt"
    args = ("8400","0",logfile)

    # run the command, waiting until it has finished
    print("Running command: {}".format(cmd+args))
    utopiaHub = subprocess.run(cmd + args, cwd=bindir, shell=False,
                               stdin=subprocess.DEVNULL)#, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print("Result: {}".format(utopiaHub.stdout))
    return utopiaHub

if __name__=="__main__":
    run()


import subprocess
import os

def run(**kwargs):
    pydir = os.path.dirname(os.path.abspath(__file__)) # mindaffectBCI/examples/acquisation/utopia_eego.py
    bindir = pydir #os.path.join(pydir,'..','..','..','bin') # ../../bin/

    # command to run the eego driver
    cmd = ("eego2utopia",)
    # args to pass to the exe
    args = () #"8400","0","../logs/mindaffectBCI.txt")

    # run the command, waiting until it has finished
    print("Running command: {}".format(cmd+args))
    # TODO []: shell=True shouldn't be needed....
    eego2utopia = subprocess.run(cmd + args, shell=True, cwd=bindir, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print("Result: {}".format(eego2utopia.stdout))
    return eego2utopia

if __name__=="__main__":
    run()


import numpy as np
from mindaffectBCI.decoder.offline.load_mindaffectBCI  import load_mindaffectBCI
from mindaffectBCI.utopiaclient import  DataPacket
from mindaffectBCI.decoder.model_fitting import MultiCCA
from mindaffectBCI.decoder.utils import window_axis
import matplotlib.pyplot as plt
import glob


def triggerPlot(filename=None, evtlabs=('0','1'), tau_ms=400, offset_ms=-50, max_samp=6000, stopband=(.1,45,'bandpass'), fs_out=250, **kwargs):
    """make a set of visualizations of the stimulus->measurement time-lock

    Args:
        filename ([type], optional): The name of the file to load. Defaults to None.
        evtlabs (tuple, optional): The event coding to use for how stimulus changes map into trigger inputs. Defaults to ('0','1').
        tau_ms (int, optional): The duration of the stimulus response. Defaults to 400.
        offset_ms (int, optional): Offset of the start of the stimulus response w.r.t. the trigger time. Defaults to -50.
        max_samp (int, optional): Limit in the number of samples for each trial. Defaults to 6000.
        stopband (tuple, optional): Temporal filter used to pre-process the EEG. Defaults to (.1,45,'bandpass').
        fs_out (int, optional): Sample rate of the EEG used for analysis. Defaults to 250.
    """   
    import glob
    import os
    if filename is None or filename == '-':
        # default to last log file if not given
        files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/mindaffectBCI*.txt')) # * means all if need specific format then *.csv
        filename = max(files, key=os.path.getctime)
    else:
        files = glob.glob(os.path.expanduser(filename))
        filename = max(files, key=os.path.getctime)    

    X, Y, coords = load_mindaffectBCI(filename, stopband=stopband, fs_out=fs_out, **kwargs)
    # size limit...
    if max_samp is not None and X.shape[1] > max_samp:
        X=X[:,:max_samp,...]
        Y=Y[:,:max_samp,...]
    X[...,:-1] = X[...,:-1] - np.mean(X[...,:-1],axis=-2,keepdims=True) # offset remove
    fs = coords[-2]['fs']
    print("EEG: X({}){} @{}Hz".format([c['name'] for c in coords],X.shape,coords[1]['fs']))
    print("STIMULUS: Y({}){}".format([c['name'] for c in coords[:-1]]+['output'],Y.shape))

    plt.figure(1)
    plt.clf()
    for i in range(min(X.shape[0],3)):
        plt.subplot(3,1,i+1)
        #plt.imshow(X[0,...].T,aspect='auto',label='X',extent=[0,X.shape[-2],0,X.shape[-1]]);
        for c in range(X.shape[-1]):
            tmp = X[i,...,c]
            tmp = (tmp - np.mean(tmp.ravel())) / max(1,np.std(tmp.ravel()))
            plt.plot(tmp+2*c,label='X{}'.format(c))
        plt.plot(Y[i,...,0],'k',label='Y')
        plt.title('Trl {}'.format(i))
    plt.legend()
    plt.suptitle('{}\nFirst trials data vs. stimulus'.format(filename))
    plt.show(block=False)
    plt.pause(.1) # allow full redraw

    print("training model")
    tau = int(fs*tau_ms/1000.0)
    # BODGE: for speed only use first 5 trials!
    # BODGE: reg with 1e-5 so only the strong channels are used..
    clsfr = MultiCCA(evtlabs=evtlabs,tau=tau,rank=1,reg=(1e-2,None)).fit(X[:5,:1000,...],Y[:5,:1000,...])

    # get the event-coded version of Y
    Ye = clsfr.stim2event(Y)

    print("applying spatial filter")
    W = clsfr.W_[0,0,...] # (d,)
    print("W={}".format(W))

    # slice the data w.r.t. the stimulus triggers to generate the visualization
    offset = int(fs*offset_ms/1000.0)
    times = (np.arange(tau)+offset)*1000/fs
    wX = np.einsum("d,Ttd->Tt", W, X) # (nTrl, nSamp) apply the spatial filter

    # add as line to the trl plot:
    for i in range(min(X.shape[0],3)):
        plt.subplot(3,1,i+1)
        tmp = wX[i,...]
        tmp = (tmp - np.mean(tmp.ravel())) / max(.01,np.std(tmp.ravel()))
        plt.plot(tmp+2*(X.shape[-1]+1),label='wX')
    plt.legend()

    # model in a new figure
    plt.figure(2)
    clsfr.plot_model(fs=fs)
    plt.show(block=False)
    # allow full re-draw
    plt.pause(.5)

    # slice out the responses for the trigger stimulus
    print('slicing data')
    wXe = window_axis(wX, winsz=tau, axis=-1) # (nTrl, nSamp-tau, tau)
    # N.B. negative offset here as negative shift of Y is same as positive shift of X
    Y_true = Ye[...,-offset:wXe.shape[1]-offset,0,0] # (nTrl, nSamp-tau)
    wXeY = wXe[Y_true>0,:] # [ep,tau]

    # get a samp#, timestamp dataset
    trl_ts = coords[0]['trl_ts'][:,:wXe.shape[1]]   # (nTrl, nSamp-tau)
    trl_idx = coords[0]['trl_idx'][:,:wXe.shape[1]] # (nTrl, nSamp-tau)
    samp2ms = np.median((np.diff(trl_ts,axis=-1)/np.maximum(1,np.diff(trl_idx,axis=-1))).ravel())
    samp2ms = 1000/fs #samp2ms*.99
    print("samp2ms={}".format(samp2ms))
    lin_trl_ts = trl_idx*samp2ms # linear time-stamp assuming constant sample rate
    ts_err = trl_ts - lin_trl_ts # (tr, nSamp-tau) error btw. linear time-stamp and observed
    ts_err = ts_err - ts_err[0,0]
    ts_errY = ts_err[Y_true>0] # (ep,) slice out the high-stimulus info (in seconds)

    print('generating plot')
    plt.figure(3)
    mu = np.median(wXeY.ravel())
    scale = np.median( np.abs(wXeY.ravel()-mu) )
    fig, ax = plt.subplots()
    plt.imshow(wXeY.T,origin='lower',aspect='auto',extent=[0,wXeY.shape[0],times[0],times[-1]])
    plt.clim(mu-scale,mu+scale)
    plt.set_cmap('jet')#'gray')#'nipy_spectral')
    plt.colorbar()
    plt.ylabel('time (ms)')
    plt.xlabel('Epoch')
    plt.title('{} locked {} trials {:4.1f}min data\n{}'.format(evtlabs[0],X.shape[0],(trl_ts[-1,-1]-trl_ts[0,0])/1000/60,filename[-50:]))
    plt.grid()
    #plt.colorbar()

    # if X.shape[-1]>8 :
    #     samp_cnt = X[...,-1] # sample count (%255)
    #     samp_miss = samp_cnt[...,1:] - ((samp_cnt[...,:-1]+1) % 256)
    #     samp_miss = samp_miss[...,:X.shape[1]]
    #     samp_miss = samp_miss[Y_true>0]
    #     plt.plot(samp_miss*samp2ms + 100,'k-')

    # over-plot the error
    ax2=ax.twinx()
    plt.plot(ts_errY,'k-')
    mu2 = np.median(ts_errY.ravel())
    scale2 = np.median( np.abs(ts_errY.ravel()-mu) )
    if times[-1]*.05 < scale2*2 and scale2*2 < times[-1]*8: # make line match
        scale2 = times[-1]/2
    elif scale2*2 < times[-1]*.2: # make image smaller
        ax.set_ylim(min(times[0],mu2-2*scale2),max(times[-1],mu2+2*scale2))
    plt.ylim(mu2-1*scale2,mu2+1*scale2)
    plt.ylabel("Recieved time-stamp error vs. constant rate (ms)")
    plt.show(block=False)

    # allow full re-draw
    plt.pause(.5)

    # now plot as lines
    plt.figure(5)
    tmp = wXeY.reshape((-1,wXeY.shape[-1]))[:100,:]
    tmp = tmp - np.mean(tmp,-1,keepdims=True)
    tmp = tmp / np.std(tmp.ravel())
    plt.plot(times, tmp.T + np.arange(tmp.shape[0])[np.newaxis,:])
    plt.title('1st {} {} evt trigger epochs time-series'.format(tmp.shape[0],evtlabs[0]))

    # do a time-stamp check.
    from mindaffectBCI.decoder.timestamp_check import timestampPlot
    plt.figure(6)
    timestampPlot(filename)

if __name__=="__main__":
    #filename="~/Desktop/trig_check/mindaffectBCI_*brainflow*.txt"
    #filename = '~/Desktop/rpi_trig/mindaffectBCI_*_201001_1859.txt'
    filename = '~/Desktop/trig_check/mindaffectBCI_*wifi*tschannel*.txt'
    #filename = '~/Desktop/trig_check/mindaffectBCI_*_khash2.txt'
    #filename=None
    filename='c:/Users/Developer/Desktop/pymindaffectBCI/logs/mindaffectBCI_*.txt'
    triggerPlot(filename, evtlabs=('re','fe'), tau_ms=400, offset_ms=-50, stopband=(0,.5), fs_out=250)#, timestamp_ch=8)




from setuptools import setup, find_packages
import glob

with open("README.rst", encoding='utf-8') as fh:
    long_description = fh.read()

setup(name='mindaffectBCI',
      version='0.9.18',
      description='The MindAffect BCI python SDK',
      long_description_content_type='text/x-rst',      
      long_description=long_description,
      url='http://github.com/mindaffect/pymindaffectBCI',
      author='Jason Farquhar',
      author_email='jason@mindaffect.nl',
      license='MIT',
      packages=['mindaffectBCI','mindaffectBCI/decoder','mindaffectBCI/decoder/offline','mindaffectBCI/examples/presentation','mindaffectBCI/examples/output','mindaffectBCI/examples/utilities','mindaffectBCI/examples/acquisation'],#,find_packages(),#
      include_package_data=True,
      package_data={'mindaffectBCI':glob.glob('mindaffectBCI/*.txt'), 'mindaffectBCI':glob.glob('mindaffectBCI/*.png'), 'mindaffectBCI.decoder':glob.glob('mindaffectBCI/decoder/*.pk')},
      data_files=[('mindaffectBCI',glob.glob('mindaffectBCI/*.png')), ('mindaffectBCI',glob.glob('mindaffectBCI/*.txt')), ('mindaffectBCI/decoder',glob.glob('mindaffectBCI/decoder/*.pk'))],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      python_requires='>=3.5',
      install_requires=['numpy>=1.0.2', 'pyglet>=1.2', 'scipy>=1.0', 'brainflow>=3.0',
'matplotlib>=3.0'],
      #entry_points={ 'console_scripts':['online_bci=mindaffectBCI.online_bci']},
      zip_safe=False)



#!/usr/bin/env python3
import numpy as np
from mindaffectBCI.decoder.UtopiaDataInterface import UtopiaDataInterface, butterfilt_and_downsample
from mindaffectBCI.utopiaclient import NewTarget, Selection, ModeChange, PredictedTargetDist, PredictedTargetProb
from mindaffectBCI.decoder.devent2stimsequence import devent2stimSequence, upsample_stimseq
from mindaffectBCI.decoder.model_fitting import BaseSequence2Sequence, MultiCCA
from mindaffectBCI.decoder.decodingSupervised import decodingSupervised
from mindaffectBCI.decoder.decodingCurveSupervised import decodingCurveSupervised, plot_decoding_curve
from mindaffectBCI.decoder.scoreOutput import dedupY0
from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, plot_summary_statistics, plot_erp
import os

PREDICTIONPLOTS = False
CALIBRATIONPLOTS = False

def get_trial_start_end(msgs, start_ts=None):
    '''get the start+end times of the trials in a utopia message stream'''
    trials = []
    keeplast = False
    for mi, m in enumerate(msgs):
        #print("msg={}".format(m))
        # process begin trail messages, N.B. after end-trial!
        if m.msgID == NewTarget.msgID:
            if start_ts is None:
                start_ts = m.timestamp
                print("NT: tr-bgn={}".format(start_ts))
            else: # treat as end of sequence + start next sequence
                trials.append((start_ts, m.timestamp))
                start_ts = m.timestamp
                print("NT: tr-end={}".format(m.timestamp))
                print("NT: tr-bgn={}".format(m.timestamp))
            
        # process the end-trial messages
        if m.msgID == Selection.msgID:
            if start_ts is not None:
                trials.append((start_ts, m.timestamp))
                start_ts = m.timestamp
                print("SL: tr-end={}".format(m.timestamp))
            else:
                print("Selection without start")
        
        if m.msgID == ModeChange.msgID:
            if start_ts is not None:
                trials.append((start_ts,m.timestamp))
                start_ts=m.timestamp
                print("MC: tr-end={}".format(m.timestamp))
            print("mod-chg={}".format(m))
            # mark this message a *not* processed so it gets to the main loop
            keeplast=True
            break
    # make list un-processed messages
    if msgs and keeplast:
        msgs = msgs[mi:]
        #print("unproc messages: {}".format(msgs))
    # return trial start/end + non-processed messages
    # N.B. start_ts is not None if trail start without end..
    return (trials, start_ts, msgs)

def getCalibration_dataset(ui):
    ''' extract a labelled dataset from the utopiaInterface, which are trilas between modechange messages '''
    # run until we get a mode change gathering training data in trials
    dataset = []
    start_ts = None
    isCalibrating = True
    while isCalibrating:

        # get new messages from utopia-hub
        newmsgs, _, _ = ui.update()
        #  print("Extact_msgs:"); print("{}".format(newmsgs))

        # incremental extract trial limits
        trials, start_ts, newmsgs = get_trial_start_end(newmsgs, start_ts)

        # extract any complete trials data/msgs
        for (bgn_ts, end_ts) in trials:
            # N.B. be sure to make a copy so isn't changed outside us..
            data = ui.extract_data_segment(bgn_ts, end_ts)
            stimulus = ui.extract_stimulus_segment(bgn_ts, end_ts)
            print("Extract trl: {}->{}: data={} stim={}".format(bgn_ts, end_ts, data.shape, stimulus.shape))
            dataset.append((data, stimulus))

        # check for end-calibration messages
        for i, m in enumerate(newmsgs):
            if m.msgID == ModeChange.msgID:
                isCalibrating = False
                # return unprocessed messages including the mode change
                #  print('cal pushback: {}'.format(newmsgs[i:]))
                ui.push_back_newmsgs(newmsgs[i:])
                break

    return dataset

def dataset_to_XY_ndarrays(dataset):
    ''' convert variable-trial length dataset with stimulus information to fixed size sklearn compatiable (X,Y) nd-arrays'''
    # get length of each trial
    trlen = [trl[0].shape[0] for trl in dataset]
    trstim = [trl[1].shape[0] for trl in dataset]
    print("Trlen: {}".format(trlen))
    print("Trstim: {}".format(trstim))
    # set array trial length to 90th percential length
    trlen = int(np.percentile(trlen, 75))
    trstim = max(20, int(np.percentile(trstim, 75)))
    # filter the trials to only be the  ones long enough to be worth processing
    dataset = [d for d in dataset if d[0].shape[0] > trlen//2 and d[1].shape[0] > trstim//2]
    if trlen == 0 or len(dataset) == 0:
        return None, None

    # map to single fixed size matrix + upsample stimulus to he EEG sample rate
    Y = np.zeros((len(dataset), trlen, 256), dtype=dataset[0][1].dtype)
    X = np.zeros((len(dataset), trlen, dataset[0][0].shape[-1]-1), dtype=dataset[0][0].dtype)  # zero-padded data, w/o time-stamps
    X_ts = np.zeros((len(dataset),trlen),dtype=int)
    Y_ts = np.zeros((len(dataset),trlen),dtype=int)
    for ti, (data, stimulus) in enumerate(dataset):
        # extract the data & remove the timestamp channel and insert into the ndarray
        # guard for slightly different sizes..
        if X.shape[1] <= data.shape[0]:
            X[ti, :, :] = data[:X.shape[1], :-1]
            X_ts[ti, :] = data[:X.shape[1], -1]
        else:  # pad end with final value
            X[ti, :data.shape[0], :] = data[:, :-1]
            X[ti, data.shape[0]:, :] = data[-1, :-1]
            X_ts[ti, :data.shape[0]] = data[:, -1]

        # upsample stimulus to the data-sample rate and insert into ndarray
        data_ts = data[:, -1]  # data timestamp per sample
        stimulus_ts = stimulus[:, -1]  # stimulus timestamp per stimulus event
        stimulus, data_i = upsample_stimseq(data_ts, stimulus[:, :-1], stimulus_ts)
        # store -- compensating for any variable trial lengths.
        if Y.shape[1] < stimulus.shape[0]: # long trial
            Y[ti, :, :] = stimulus[:Y.shape[1], :]
        else: # short trial
            Y[ti, :stimulus.shape[0], :] = stimulus
        # record stim-ts @ this data_ts
        tmp = data_i < Y.shape[1]
        Y_ts[ti,data_i[tmp]] = stimulus_ts[tmp] 

    return X, Y, X_ts, Y_ts


def strip_unused(Y):
    # strip unused outputs
    used_y = np.any(Y.reshape((-1, Y.shape[-1])), 0)
    used_y[0] = True # ensure objID=0 is always used..
    Y = Y[..., used_y]
    return Y, used_y


def doCalibrationSupervised(ui: UtopiaDataInterface, clsfr: BaseSequence2Sequence, 
                            cv=2, calfn="calibration_data.pk", fitfn="fit_data.pk", previous_dataset=None):
    ''' do a calibration phase = basically just extract  the training data and train a classifier from the utopiaInterface'''
    X = None
    Y = None
    dataset = getCalibration_dataset(ui)
    if previous_dataset is not None: # combine with the old calibration data
        dataset.extend(previous_dataset)
    if dataset:
        # convert msgs -> to nd-arrays
        X, Y, X_ts, Y_ts = dataset_to_XY_ndarrays(dataset)

        try:
            import pickle
            pickle.dump(dict(X=X, Y=Y, X_ts=X_ts, Y_ts=Y_ts, fs=ui.fs),
                        open('calibration_data.pk','wb'))
        except:
            print('Error saving cal data')

        # guard against empty training dataset
        if X is None or Y is None :
            return None, None, None
        Y, used_idx = strip_unused(Y)
        
        # now call the clsfr fit method, on the true-target info
        try:
            print("Training dataset = ({},{})".format(X.shape, Y.shape))
            cvscores = clsfr.cv_fit(X, Y, cv=cv)
            score = np.mean(cvscores['test_score'])
            print("clsfr={} => {}".format(clsfr, score))
        except:
            import traceback
            traceback.print_exc()
            return None, None, None

        decoding_curve = decodingCurveSupervised(cvscores['estimator'], nInt=(10, 10),
                                      priorsigma=(clsfr.sigma0_, clsfr.priorweight),
                                      softmaxscale=clsfr.softmaxscale_)
        # extract the final estimated performance
        #print("decoding curve {}".format(decoding_curve[1]))
        #print("score {}".format(score))
        perr = decoding_curve[1][-1] if len(decoding_curve)>1 else 1-score
        if CALIBRATIONPLOTS:
            try:
            #if True:
                import matplotlib.pyplot as plt
                plt.figure(1)
                clsfr.plot_model(fs=ui.fs)
                plt.suptitle('Factored Model')
                plt.figure(2)
                plot_decoding_curve(*decoding_curve)
                #  from analyse_datasets import debug_test_dataset
                #  debug_test_dataset(X,Y,None,fs=ui.fs)
                plt.figure(3) # plot the CCA info
                Y_true = clsfr.stim2event(Y)
                Y_true = Y_true[...,0:1,:]
                Cxx, Cxy, Cyy = updateSummaryStatistics(X,Y_true,tau=clsfr.tau)
                plot_summary_statistics(Cxx,Cxy,Cyy,clsfr.evtlabs,fs=ui.fs)
                plt.suptitle("Summary Statistics")
                try:
                    import pickle
                    pickle.dump(dict(Cxx=Cxx, Cxy=Cxy, Cyy=Cyy, evtlabs=clsfr.evtlabs, fs=ui.fs),
                                open('summary_statistics.pk','wb'))
                except:
                    print('Error saving cal data')
                plt.figure(4)
                plot_erp(Cxy,evtlabs=clsfr.evtlabs,fs=ui.fs)
                plt.suptitle("Event Related Potential (ERP)")
                plt.show(block=False)
                # save figures
                logsdir = os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/')
                plt.figure(1)
                plt.savefig(os.path.join(logsdir,'model_{}.png'.format(uname)))
                plt.figure(2)
                plt.savefig(os.path.join(logsdir,'decoding_curve_{}.png'.format(uname)))
                plt.figure(3)
                plt.savefig(os.path.join(logsdir,'summary_statistics_{}.png'.format(uname)))
                plt.figure(4)
                plt.savefig(os.path.join(logsdir,'erp_{}.png'.format(uname)))
            except:
                pass

        # send message with calibration performance score
        ui.sendMessage(PredictedTargetProb(ui.stimulus_timestamp, 0, perr))
        
    return dataset, X, Y


def doPrediction(clsfr: BaseSequence2Sequence, data, stimulus, prev_stimulus=None):
    ''' given the current trials data, apply the classifier and decoder to make target predictions '''
    X = data[:, :-1]
    X_ts = data[:, -1]
    Y = stimulus[:, :-1]
    Y_ts = stimulus[:, -1]
    if X_ts.size == 0 or Y_ts.size == 0: # fast path empty inputs
        return None
    # strip outputs that we don't use, to save compute time
    Y, used_idx = strip_unused(Y)
    # strip the true target info if it's a copy, so it doesn't mess up Py computation
    Y = dedupY0(Y, zerodup=False, yfeatdim=False)
    # up-sample Y to the match the rate of X
    # TODO[]: should this happen in the data-interface?
    Y, _ = upsample_stimseq(X_ts, Y, Y_ts)
    # predict on X,Y without the time-stamp info
    Fy_1 = clsfr.predict(X, Y, prevY=prev_stimulus)
    # map-back to 256
    Fy = np.zeros(Fy_1.shape[:-1]+(256,),dtype=Fy_1.dtype)
    Fy[..., used_idx] = Fy_1
    return Fy


def combine_Ptgt(pvals_objIDs):
    pvals = [p[0] for p in pvals_objIDs] 
    objIDs = [p[1] for p in pvals_objIDs]
    if not all(np.isequal(objIDs[0], oi) for oi in objIDs):
        print("Warning combination only supported for fixed output set currently!")
        return pvals[-1], objIDs[-1]
    pvals = np.hstack(pvals)  # (nBlk,nObj)
    # coorected combination
    Ptgt = softmax(np.sum(np.log(pvals))/np.sqrt(pvals.shape[0]))
    return Ptgt, objIDs


def send_prediction(ui: UtopiaDataInterface, Ptgt, used_idx=None, timestamp=-1):
    #print(" Pred= used_idx:{} ptgt:{}".format(used_idx,Ptgt))
    # N.B. for network efficiency, only send for non-zero probability outputs
    nonzero_idx = np.flatnonzero(Ptgt)
    # ensure a least one entry
    if nonzero_idx.size == 0: 
        nonzero_idx = [0]
    Ptgt = Ptgt[nonzero_idx]
    if used_idx is None:
        used_idx = nonzero_idx
    else:
        if np.issubdtype(used_idx.dtype, np.bool): # logical->index
            used_idx = np.flatnonzero(used_idx)
        used_idx = used_idx[nonzero_idx]
    #  print(" Pred= used_idx:{} ptgt:{}".format(used_idx,Ptgt))
    # send the prediction messages, PredictedTargetProb, PredictedTargetDist
    y_est_idx = np.argmax(Ptgt, axis=-1)
    # most likely target and the chance that it is wrong
    if Ptgt[y_est_idx] == 1.0 :
        print("P==1?") 
    else:
        ptp = PredictedTargetProb(timestamp, used_idx[y_est_idx], 1-Ptgt[y_est_idx])
        print(" Pred= {}".format(ptp))
        ui.sendMessage(ptp)
    # distribution over all *non-zero* targets
    ui.sendMessage(PredictedTargetDist(timestamp, used_idx, Ptgt))
    

def doPredictionStatic(ui: UtopiaDataInterface, clsfr: BaseSequence2Sequence, model_apply_type='trial', timeout_ms=None, block_step_ms=100, maxDecisLen_ms=8000):
    ''' do the prediction stage = basically extract data/msgs from trial start and generate a prediction from them '''

    if not clsfr.is_fitted():
        print("Warning: trying to predict without training classifier!")
        return

    # TODO []: Block based prediction is slightly slower?  Why?
    if timeout_ms is None:
        timeout_ms = block_step_ms
    
    # start of the data block to apply the model to
    block_start_ts = ui.data_timestamp
    overlap_samp = clsfr.tau
    overlap_ms = overlap_samp * 1000 / ui.fs
    maxDecisLen_samp = int(maxDecisLen_ms * ui.fs / 1000)
    Fy = None # (1,nSamp,nY):float score for each output for each sample
    trial_start_ts = None
    isPredicting = True
    # run until we get a mode change gathering training data in trials
    while isPredicting:
        # get new messages from utopia-hub
        newmsgs, ndata, nstim = ui.update(timeout_ms=timeout_ms,mintime_ms=timeout_ms//2)

        # TODO[]: Fix to not re-process the same data if no new stim to be processed..
        if len(newmsgs) == 0 and nstim == 0 and ndata == 0:
            continue

        # get the timestamp for the last data which it is valid to apply the model to,
        # that is where have enough data to include a complete response for this stimulus
        # Note: can't just use last data, incase stimuli are lagged w.r.t. data
        # also, prevents processing data for which are not stimulus events to compare with
        valid_end_ts = min(ui.stimulus_timestamp + overlap_ms, ui.data_timestamp)

        # incremental extract trial limits
        otrial_start_ts = trial_start_ts
        trials, trial_start_ts, newmsgs = get_trial_start_end(newmsgs, trial_start_ts)

        # change in trial-start -> end-of-trial / start new trial detected
        if not trial_start_ts == otrial_start_ts:
            print("New trial! tr_start={}".format(trial_start_ts))
            if PREDICTIONPLOTS and Fy is not None:
                try:
                    import matplotlib.pyplot as plt
                    plt.figure(10)
                    plt.cla()
                    print("Fy={}".format(Fy.shape))
                    plt.plot(np.cumsum(Fy[0,...],-2))
                    plt.show(block=False)
                except:
                    pass
            Fy = None
            block_start_ts = trial_start_ts

        # compute the start/end of the segement to apply the model to
        if model_apply_type == 'trial':
            # apply the model to all available data from trial start
            block_start_ts = trial_start_ts
            if block_start_ts is not None and block_start_ts + block_step_ms + overlap_ms < valid_end_ts:
                block_end_ts = valid_end_ts
            else:
                block_end_ts = None
            # limit the trial size
            if block_start_ts is not None:
                block_start_ts = max( block_start_ts, valid_end_ts - maxDecisLen_ms)
        else:
            # check if enough data to apply the model
            if block_start_ts is not None and block_start_ts + block_step_ms + overlap_ms < valid_end_ts:
                # got enough data to process this block
                block_end_ts = valid_end_ts
            else:
                # not enough yet -> clear the end to indicate dont apply the model
                block_end_ts = None

        # if we have a valid block to apply the model do
        if block_start_ts is not None and block_end_ts is not None:

            # extract and apply to this block
            print("Extract block: {}->{} = {}ms".format(block_start_ts, block_end_ts, block_end_ts-block_start_ts))
            data = ui.extract_data_segment(block_start_ts, block_end_ts)
            stimulus = ui.extract_stimulus_segment(block_start_ts, block_end_ts)
            # skip if no data/stimulus to process
            if data.size == 0 or stimulus.size == 0:
                continue

            print('got: data {}->{} ({})  stimulus {}->{} ({}>0)'.format(data[0, -1], data[-1, -1], data.shape[0],
                                                              stimulus[0, -1], stimulus[-1, -1], np.sum(stimulus[:,0])))
            if model_apply_type == 'block':
                # update the start point for the next block
                # start next block at overlap before the end of this blocks data
                # so have sample accurate predictions, with no missing data, and no overlaps
                block_start_ts = data[-overlap_samp+1, -1]  # ~= block_end_ts - overlap_ms +1-sample
                bend = block_start_ts + block_step_ms + overlap_ms
                print("next block {}->{}: in {}ms".format(block_start_ts, bend, bend - ui.data_timestamp))

            # get predictions for this data block
            block_Fy = doPrediction(clsfr, data, stimulus)
            # strip predictions from the overlap period
            block_Fy = block_Fy[..., :-overlap_samp, :]

            # if got valid predictions...
            if block_Fy is not None:
                # accumulate or store the predictions
                if model_apply_type == 'trial':
                    Fy = block_Fy
                elif model_apply_type == 'block':  # accumulate blocks in the trial
                    if Fy is None:  # restart accumulation
                        Fy = block_Fy
                    else:
                        Fy = np.append(Fy, block_Fy, -2)
                # limit the trial length
                if maxDecisLen_ms > 0 and Fy.shape[-2] > maxDecisLen_samp:
                    print("limit trial length {} -> {}".format(Fy.shape[-2], maxDecisLen_samp))
                    Fy = Fy[..., -maxDecisLen_samp:, :]

                # send prediction event
                # only process the used-subset
                used_idx = np.any(Fy.reshape((-1, Fy.shape[-1])), 0)
                used_idx[0] = True # force include 0
                # map to probabilities, including the prior over sigma!
                Ptgt = clsfr.decode_proba(Fy[...,used_idx])
                #  _, _, Ptgt, _, _ = decodingSupervised(Fy[...,used_idx])
                # BODGE: only  use the last (most data?) prediction...
                Ptgt = Ptgt[-1, -1, :]
                # send prediction with last recieved stimulus_event timestamp
                print("Fy={} Yest={} Perr={}".format(Fy.shape, np.argmax(Ptgt), 1-np.max(Ptgt)))

                send_prediction(ui, Ptgt, used_idx=used_idx, timestamp=ui.stimulus_timestamp)
            
        # check for end-prediction messages
        for i,m in enumerate(newmsgs):
            if m.msgID == ModeChange.msgID:
                isPredicting = False
                # return unprocessed messages to stack. Q: why i+1?
                ui.push_back_newmsgs(newmsgs[i:])

def run(ui: UtopiaDataInterface=None, clsfr: BaseSequence2Sequence=None, msg_timeout_ms: float=100, 
        host:str=None, datafile:str=None,
        tau_ms:float=450, offset_ms:float=0, out_fs:float=100, evtlabs=None, 
        stopband=((45,65),(5.5,25,'bandpass')), ftype='butter', order:int=6, cv:int=5, 
        calplots:bool=False, predplots:bool=False, label:str=None, **kwargs):
    """ run the main decoder processing loop

    Args:
        ui (UtopiaDataInterface, optional): The utopia data interface class. Defaults to None.
        clsfr (BaseSequence2Sequence, optional): the classifer to use when model fitting. Defaults to None.
        msg_timeout_ms (float, optional): timeout for getting new messages from the data-interface. Defaults to 100.
        host (str, optional): hostname for the utopia hub. Defaults to None.
        tau_ms (float, optional): length of the stimulus response. Defaults to 400.
        offset_ms (float, optiona): offset in ms to shift the analysis window. Use to compensate for response lag.  Defaults to 0.
        stopband (tuple, optional): temporal filter specification for `UtopiaDataInterface.butterfilt_and_downsample`. Defaults to ((45,65),(5.5,25,'bandpass'))
        ftype (str, optional): type of temporal filter to use.  Defaults to 'butter'.
        order (int, optional): order of temporal filter to use.  Defaults to 6.
        out_fs (float, optional): sample rate after the pre-processor. Defaults to 100.
        evtlabs (tuple, optional): the brain event coding to use.  Defaults to None.
        calplots (bool, optional): flag if we make plots after calibration. Defaults to False.
        predplots (bool, optional): flag if we make plots after each prediction trial. Defaults to False.
    """
    global CALIBRATIONPLOTS, PREDICTIONPLOTS
    CALIBRATIONPLOTS = calplots
    PREDICTIONPLOTS = predplots
    try :
        import matplotlib.pyplot as plt
        guiplots=True
    except:
        guiplots=False

    # setup the saving label
    global uname
    from datetime import datetime 
    uname = datetime.now().strftime("%y%m%d_%H%M")
    if label is not None: # include label as prefix
        uname = "{}_{}".format(label,uname)

    # create data interface with bandpass and downsampling pre-processor, running about 10hz updates
    if ui is None:
        try:
            from  scipy.signal import butter
            ppfn = butterfilt_and_downsample(order=order, stopband=stopband, fs_out=out_fs, ftype=ftype)
        except: # load filter from file
            print("Warning: stopband specification *ignored*, using sos_filter_coeff.pk file...")
            ppfn = butterfilt_and_downsample(stopband='sos_filter_coeff.pk', fs_out=out_fs)
        #ppfn = None
        ui = UtopiaDataInterface(data_preprocessor=ppfn,
                                 stimulus_preprocessor=None,
                                 timeout_ms=100, mintime_ms=55, clientid='decoder') # 20hz updates
    ui.connect(host=host, queryifhostnotfound=False)
    # use a multi-cca for the model-fitting
    if clsfr is None:
        if isinstance(evtlabs,str): # decode string coded spec
            evtlabs = evtlabs.split(',')
        clsfr = MultiCCA(tau=int(out_fs*tau_ms/1000), evtlabs=evtlabs, offset=int(out_fs*offset_ms/1000))
        print('clsfr={}'.format(clsfr))
    
    calibration_dataset = None
    current_mode = "idle"
    # clean shutdown when told shutdown
    while current_mode.lower != "shutdown".lower():

        if  current_mode.lower() in ("calibration.supervised","calibrate.supervised"):
            calibration_dataset, _, _ = doCalibrationSupervised(ui, clsfr, cv=cv, previous_dataset=calibration_dataset)
                
        elif current_mode.lower() in ("prediction.static","predict.static"):
            doPredictionStatic(ui, clsfr)

        elif current_mode.lower() in ("reset"):
            calibration_dataset = None

        # check for new mode-messages
        newmsgs, nsamp, nstim = ui.update()

        # update the system mode
        current_mode = "idle"
        for i, m in enumerate(newmsgs):
            if m.msgID == ModeChange.msgID:
                current_mode = m.newmode
                print("\nNew Mode: {}".format(current_mode))
                ui.push_back_newmsgs(newmsgs[i+1:])
                # stop processing messages
                break
        
        # BODGE: re-draw plots so they are interactive.
        if guiplots:
            for i in plt.get_fignums():
                plt.figure(i).canvas.flush_events()

def mainloop(*args,**kwargs):
    run(*args,**kwargs)

if  __name__ == "__main__":
    import argparse
    import json
    parser = argparse.ArgumentParser()
    parser.add_argument('--host',type=str, help='address (IP) of the utopia-hub', default=None)
    parser.add_argument('--out_fs',type=int, help='output sample rate', default=100)
    parser.add_argument('--tau_ms',type=float, help='output sample rate', default=450)
    parser.add_argument('--evtlabs', type=str, help='comma separated list of stimulus even types to use', default='re,fe')
    parser.add_argument('--stopband',type=json.loads, help='set of notch filters to apply to the data before analysis', default=((45,65),(5.5,25,'bandpass')))
    parser.add_argument('--cv',type=int, help='number cross validation folds', default=5)
    parser.add_argument('--predplots', action='store_true', help='flag make decoding plots are prediction time')
    parser.add_argument('--calplots', action='store_false', help='turn OFF model and decoding plots after calibration')
    parser.add_argument('--savefile', type=str, help='run decoder using this file as the proxy data source', default=None)
    parser.add_argument('--savefile_fs', type=float, help='effective sample rate for the save file', default=None)

    args = parser.parse_args()

    if args.savefile is not None or True:#
        #setattr(args,'savefile',"~/utopia/java/messagelib/UtopiaMessages_.log")
        #setattr(args,'savefile',"~/utopia/java/utopia2ft/UtopiaMessages_*1700.log")
        #setattr(args,'savefile',"~/Downloads/jason/UtopiaMessages_200923_1749_*.log")
        setattr(args,'savefile','~/Desktop/mark/mindaffectBCI*.txt')
        #setattr(args,'out_fs',100)
        #setattr(args,'savefile_fs',200)
        #setattr(args,'cv',5)
        from mindaffectBCI.decoder.FileProxyHub import FileProxyHub
        U = FileProxyHub(args.savefile,use_server_ts=True)
        ppfn = butterfilt_and_downsample(order=6, stopband=args.stopband, fs_out=args.out_fs, ftype='butter')
        ui = UtopiaDataInterface(data_preprocessor=ppfn,
                                 stimulus_preprocessor=None,
                                 timeout_ms=100, mintime_ms=0, U=U, fs=args.savefile_fs, clientid='decoder') # 20hz updates
        # add the file-proxy ui as input argument
        setattr(args,'ui',ui)
    
    running=True
    nCrash = 0
    run(**vars(args))
    while running and nCrash < 10:
        try:
            run(**vars(args))
            # stop restarting if normal terminate
            running=False
        except KeyboardInterrupt:
            # stop running if keyboard interrrupt
            running=False
        except Exception as ex:
            print("Error running mainloop"+ str(ex))
            nCrash = nCrash + 1
            pass



import os
from glob import glob
import numpy as np
from scipy.io import loadmat, whosmat
from mindaffectBCI.decoder.utils import block_randomize, butter_sosfilt, window_axis


def argsort(l):
    return sorted(range(len(l)), key=l.__getitem__)

def squeeze(v):
    while v.size == 1 and v.ndim > 0:
        v = v[0]
    return v

books = ['journey','20000']
subids_attended_book=[None, #0
                      'journey','journey','20000', 'journey', 'journey', '20000', '20000', '20000', '20000', '20000', # 1-10 
                      'journey','journey','20000', 'journey', '20000', 'journey', '20000', '20000', '20000', 'journey',     # 11-20
                      '20000', 'journey', 'journey', '20000', '20000', 'journey', '20000', 'journey', 'journey', 'journey', # 21-30
                      '20000', '20000', 'journey'] # 31-33
subids_journey=[1,2, 4, 5, 11, 12, 14, 16, 20, 22, 23, 26, 28, 29, 30, 33]
subids_20000=[3, 6, 7, 8, 9, 10, 13, 15, 17, 18, 19, 21, 24, 25, 27, 31, 32]

# 128ch biosemi+10-10
ch_names=('Cz','a2','CPz','a4','a5','a6','a7','a8','a9','PO7','a11','a12','a13','a14','O1','a16','a17','a18','Pz','a20','POz','a22','Oz','a24','Iz','a26','a27','O2','a29','a30','a31','a32','b1','b2','b3','b4','b5','b6','PO8','b8','b9','b10','P8','b12','b13','TP8','b15','b16','b17','b18','b19','C2','b2','C4','b23','C6','b25','T8','FT8','b28','b29','b30','b31','b32','c1','c2','c3','c4','c5','c6','F8','AF8','c9','c10','c11','c12','c13','c14','c15','Fp2','Fpz','c18','AFz','c20','Fz','c22','FCz','c24','c25','c26','c27','c28','Fp1','AF7','c31','c32','d1','d2','d3','d4','d5','d6','F7','FT7','d9','d10','d11','d12','d13','C1','d15','d16','d17','d18','C3','d20','C5','d22','T7','TP7','d25','d26','d27','d28','d29','d30','P7','d32')

def load_cocktail(datadir, sessdir=None, sessfn=None, fs_out=60, stopband=((45,65),(0,5),(25,-1)), verb=0, trlen_ms=None, subtriallen=10):

    # load the data file
    Xfn = os.path.expanduser(datadir)
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = Xfn if os.path.isdir(Xfn) else os.path.dirname(Xfn)
    stimdir = os.path.join(sessdir,'..','..','Stimuli','Envelopes')
    
    runfns = glob(os.path.join(sessdir, '*.mat'))

    # extract subId (to get attended stream)
    subid = int(sessdir.split("Subject")[1].split("_")[0])
    attended_book = subids_attended_book[subid]
    # extract run id
    runid = [int(f.split("Run")[1].split(".mat")[0]) for f in runfns]
    # sort into numeric order
    sorted_id = argsort(runid)
    runfns = [(runid[i], runfns[i]) for i in sorted_id]

    # load the raw EEG data
    data = [None]*len(runid)
    stim = [None]*len(runid)
    print("Run:", end='')
    for i, (ri, rf) in enumerate(runfns):
        print("{} ".format(ri), end='', flush=True)
        data[i] = loadmat(rf)
        # load
        stim[i] = [loadmat(os.path.join(stimdir,book,"{}_{}_env.mat".format(book,ri))) for book in books]
    # make a label list for the trials
    lab = [books.index(attended_book)]*len(data)
    fs = squeeze(data[0]['fs'])

    if not all(squeeze(d['fs']) == fs for d in data):
        raise ValueError("Different sample rates in different runs")
    #if not all(d['fsEnv'] == fs for d in stim):
    #    raise valueError("Different samples rates in between EEG and Envelope")
        
    # make the X and Y arrays
    X0 = data[0]['eegData']
    Y0 = stim[0][0]['envelope']
    nSamp = min(X0.shape[0], Y0.shape[0])
    d = X0.shape[1]
    e = Y0.shape[1]
    X = np.zeros((len(runid), nSamp, d), dtype='float32')
    Y = np.zeros((len(runid), nSamp, 1+len(stim[0]), e), dtype='float32') #  (nTr,nSamp,nY,e)
    for ti,(d,s) in enumerate(zip(data, stim)):
        X[ti, :, :] = d['eegData'][:nSamp, :]
        Y[ti, :, 0, :] = s[lab[ti]]['envelope'][:nSamp, :] # objID==0 is attended
        for si,ss in enumerate(s): # all possible stimuli
            Y[ti, :, si+1, :] = ss['envelope'][:nSamp, :]
    
    print("X={}".format(X.shape), flush=True)

    # preprocess -> spectral filter, in continuous time!
    if stopband is not None:
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs)

    # preprocess -> downsample
    resamprate = int(fs/fs_out)
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs, fs_out, resamprate))
        X = X[:, ::resamprate, :] # decimate X (trl, samp, d)
        Y = Y[:, ::resamprate, ...] # decimate Y (trl, samp, y, e)
        fs = fs/resamprate

    nsubtrials = X.shape[1]/fs/subtriallen
    if nsubtrials > 1:
        winsz = int(X.shape[1]//nsubtrials)
        print('{} subtrials -> winsz={}'.format(nsubtrials,winsz))
        # slice into sub-trials
        X = window_axis(X,axis=1,winsz=winsz,step=winsz) # (trl,win,samp,d)
        Y = window_axis(Y,axis=1,winsz=winsz,step=winsz) # (trl,win,samp,nY)
        # concatenate windows into trial dim
        X = X.reshape((X.shape[0]*X.shape[1],)+X.shape[2:])
        Y = Y.reshape((Y.shape[0]*Y.shape[1],)+Y.shape[2:])
        print("X={}".format(X.shape))

    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial'}
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.arange(X.shape[1])/fs, \
                 'fs':fs}
    coords[2] = {'name':'channel','coords':ch_names}
    # return data + metadata
    return (X, Y, coords)

def testcase():
    datadir= '~/data/bci/external_data/dryad/Cocktail Party'
    datadir= 'D://external_data/dryad/Cocktail Party'
    sessdir= 'EEG/Subject1'
    sessfn = ''
    X,Y,coords=load_cocktail(datadir, sessdir, sessfn, subtriallen=15)

    from analyse_datasets import debug_test_dataset
    debug_test_dataset(X, Y, coords, tau_ms=500, rank=4, evtlabs=None)
    
    
if __name__=='__main__':
    testcase()



import argparse
import numpy as np
from mindaffectBCI import utopiaclient 
from time import time, sleep

LOGINTERVAL_S = 3
t0=None
nextLogTime=None
def printLog(nSamp, nBlock):
    ''' textual logging of the data arrivals etc.'''
    global t0, nextLogTime
    t = time()
    if t0 is None:
        t0 = t
    if nextLogTime is None:
        nextLogTime = t
    if t > nextLogTime:
        elapsed = time()-t0
        print("%d %d %f %f (samp,blk,s,hz)"%(nSamp, nBlock, elapsed, nSamp/elapsed), flush=True)
        nextLogTime = t +LOGINTERVAL_S


def parse_args():
    parser = argparse.ArgumentParser()
    # use docs to check which parameters are required for specific board, e.g. for Cyton - set serial port
    parser.add_argument('--host', type=str, help='host name for the utopia hub', required=False, default=None)
    parser.add_argument('--nch', type=str, help='number of simulated channels', required=False, default=4)
    parser.add_argument('--fs', type=str, help='simulated channels sample rate', required=False, default=200)
    args = parser.parse_args()
    return args


client = None
def run(host=None, nch: int=4, fs: float=200, packet_size: int=10):
    """run a simple fake-data stream with gaussian noise channels

    Args:
        host ([str], optional): address for the utopia hub. Defaults to None.
        nch (int, optional): number of simulated channels. Defaults to 4.
        fs (int, optional): simulated data sample rate. Defaults to 200.
        packet_size (int, optional): number channels to put in each utopia-hub datapacket. Defaults to 10.
    """    
    global client
    # connect to the utopia client
    client = utopiaclient.UtopiaClient()
    client.autoconnect(host)
    # don't subscribe to anything
    client.sendMessage(utopiaclient.Subscribe(None, ""))
    print("Putting header. {} ch @ {} Hz".format(nch,fs))
    client.sendMessage(utopiaclient.DataHeader(None, nch, fs, ""))

    # setup the ERP injection trigger listener
    import socket
    import struct
    # Create a TCP/IP socket
    trigger_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    trigger_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    trigger_sock.bind(("0.0.0.0", 8300))
    trigger_sock.setblocking(False)  # set into non-blocking mode

    nSamp = 0
    nPacket = 0
    t0 = client.getTimeStamp()
    sample_interval = 1000 / fs
    data = np.zeros((packet_size, nch), dtype=np.float32)
    while True:

        # generate a packets worth of data in 'real-time' so attach triggers
        for si in range(packet_size):
            # limit the sample rate..
            sleep(max(0, (t0 + nSamp*sample_interval - client.getTimeStamp()) / 1000))
            # generate random data 
            data[si, :] = np.random.standard_normal((nch,))
            nSamp = nSamp + 1
            # check for trigger input
            try:
                trig, addr = trigger_sock.recvfrom(1024)
                # decode the payload
                trig = struct.unpack('f', trig[:4]) if len(trig) == 4 else int(trig[0]*5)
                # add to the raw data
                data[si, -1] = data[si, -1] + trig
                print('t', end='', flush=True)
            except socket.error as ex:
                pass

        # forward to the utopia client
        nPacket = nPacket + 1
        client.sendMessage(utopiaclient.DataPacket(client.getTimeStamp(), data))

        printLog(nSamp, nPacket)        


if __name__ == "__main__":
    args = parse_args()
    run(**vars(args))


#!/usr/bin/env python3
import numpy as np
from mindaffectBCI.decoder.UtopiaDataInterface import UtopiaDataInterface, stim2eventfilt, butterfilt_and_downsample


def erpViewer(ui: UtopiaDataInterface, timeout_ms: float=np.inf, tau_ms: float=500,
              offset_ms=(-15, 0), evtlabs=None, ch_names=None, nstimulus_events: int=600, rank=3, center=True):
    ''' simple sig-viewer using the ring-buffer for testing '''
    import matplotlib.pyplot as plt

    if evtlabs is None:
        evtlabs = ('re', 'fe')

    # initialize the filter from raw stimulus to brain-trigger events
    stimfilt = stim2eventfilt(evtlabs)

    ui.update()

    # compute the size of the erp slice
    irf_range_ms = (offset_ms[0], tau_ms+offset_ms[1])
    irflen_ms = irf_range_ms[1]-irf_range_ms[0]
    irflen_samp = int(np.ceil(irflen_ms * ui.fs / 1000))
    irf_times = np.linspace(irf_range_ms[0], irf_range_ms[1], irflen_samp)

    # initialize the arrays to hold the stimulus responses, and their label info, and a current cursor
    irf = np.zeros((nstimulus_events, irflen_samp, ui.data_ringbuffer.shape[-1] - 1)) # (nERP,irflen,d)
    irf_lab = np.zeros((nstimulus_events, len(evtlabs)), dtype=int)  # (nErp,nstim)
    cursor = 0

    # initialize the plot window
    fig = plt.figure(1)
    fig.clear()

    # main spec, inc titles
    outer_grid = fig.add_gridspec(nrows=2, ncols=2, height_ratios=[1,8])

    # right-sub-spec for the ERPs
    plt.figtext(.25,.9,'ERPs',ha='center')
    # get grid-spec to layout the plots, 1 for spatial, 1 for temporal
    gs = outer_grid[-1,0].subgridspec(nrows=len(evtlabs), ncols=1)
    erp_ax = [None for j in range(len(evtlabs))]
    erp_lines = [None for j in range(len(evtlabs))]
    erp_ax[-1] = fig.add_subplot(gs[-1,0])
    for ei,lab in enumerate(evtlabs):
        if ei==len(evtlabs)-1:
            ax = erp_ax[ei]
            ax.set_xlabel("Time (ms)")
            ax.set_ylabel("Space")
        else:
            ax = fig.add_subplot(gs[ei,0], sharex=erp_ax[-1], sharey=erp_ax[-1])
            ax.tick_params(labelbottom=False)
        erp_lines[ei] = ax.imshow(irf[ei,:,:].T,origin='lower',aspect='auto',extent=(irf_times[0],irf_times[-1],0,irf.shape[-1]))
        ax.set_title(lab)
        erp_ax[ei] = ax

    # left-sub-spec for the decomposition
    plt.figtext(.75,.9,'SVD Decomposition',ha='center')

    # get grid-spec to layout the plots, 1 for temporal, rank for spatial
    gs =  outer_grid[-1,1].subgridspec(nrows=len(evtlabs), ncols=rank+1, width_ratios=(4,)+(1,)*rank)
    spatial_ax = [[ None for i in range(rank)] for j in range(len(evtlabs))]
    spatial_lines = [[ None for i in range(rank)] for j in range(len(evtlabs))]
    temporal_ax = [None for j in range(len(evtlabs))]
    temporal_lines = [None for j in range(len(evtlabs))]
    # make the bottom 2 axs already so can share their limits.
    temporal_ax[-1] = fig.add_subplot(gs[-1, 0])
    spatial_ax[-1][0] = fig.add_subplot(gs[-1, 1])
    for ei, lab in enumerate(evtlabs):
        # single temporal plot for all ranks
        if ei == len(evtlabs)-1 and ri==0: # tick-plot
            ax = temporal_ax[ei]
            ax.set_xlabel("Time (ms)")
        else:
            ax = fig.add_subplot(gs[ei, 0], sharex=temporal_ax[-1], sharey=temporal_ax[-1])
            ax.tick_params(labelbottom=False)
        ax.set_title("{}".format(lab))
        ax.grid(True)
        ax.autoscale(axis='y')
        temporal_ax[ei] = ax
        # component range
        temporal_lines[ei] = ax.plot(irf_times, np.ones((irf.shape[-2], rank)))
        # TODO: add errorbar visualization..
        # component range
        #temporal_lines[ei].extend(ax.plot(irf_times, np.ones((irf.shape[-2], rank)), color=(0, 0, 0)))

        for ri in range(rank):
            # spatial-plot
            if ei == len(evtlabs)-1 and ri==0: # tick-plot
                ax = spatial_ax[ei][ri]
                ax.set_xlabel("Space")
            else:
                ax = fig.add_subplot(gs[ei, 1+ri], sharex=spatial_ax[-1][0], sharey=spatial_ax[-1][0])
                ax.tick_params(labelbottom=False, labelleft=False)

            ax.set_title("{}".format(lab))
            ax.grid(True)
            spatial_ax[ei][ri] = ax
            spatial_lines[ei][ri] = ax.plot(np.zeros((irf.shape[-1])), color=(0, 0, 0))

    # add a reset ERP button
    def reset(event):
        print('reset called')
        irf_lab.fill(0)
        cursor=0 
    from matplotlib.widgets import Button
    butax = fig.add_axes([.05,.85,.1,.1])
    breset = Button(butax,'Reset')
    breset.on_clicked(reset)


    fig.show()
    
    # start the render loop
    t0 = ui.getTimeStamp()
    pending = []  # list of stimulus events for which still waiting for data to complete
    while ui.getTimeStamp() < t0+timeout_ms:

        # exit if figure is closed..
        if not plt.fignum_exists(1):
            quit()

        # re-draw the display
        fig.canvas.draw()
        fig.canvas.flush_events()
        #plt.pause(.001)

        # Update the records
        nmsgs, ndata, nstim = ui.update(timeout_ms=100)

        # skip if no stimulus events to process
        if nstim == 0:
            continue

        # get the new raw stimulus
        rawstimulus = ui.stimulus_ringbuffer[-nstim:, :]

        # split the time-stamp and the *true-target* stimulus info
        stimulus_ts = rawstimulus[:, -1]
        # and transform to irf trigger events
        stimulus = stimfilt.transform(rawstimulus)  # (nsamp,nstim)
        # strip to only the objID==0 output
        stimulus = stimulus[:, 0, :]  # (nsamp,nE)

        # identify any new stimulus events and add to the IRF stack
        for ti in np.flatnonzero(np.any(stimulus, -1)):
            # get the trigger time-stamp
            ts = stimulus_ts[ti]
            # get the trigger data range
            data_segment = (ts+irf_range_ms[0], ts+irf_range_ms[1])
            # store in the pending list
            pending.append((stimulus[ti, :], data_segment))

        # process any completed pending stimulus segments
        new_pending = []
        for stimulus_segment in pending:
            stimulus, data_segment = stimulus_segment
            if data_segment[1] < ui.data_timestamp:
                # get the irf window
                data = ui.extract_data_segment(*data_segment)
                if data.size == 0:
                    print("didn't get any valid data!!!: data_seg={}".format(*data_segment))
                    continue
                # insert into the running buffer, w/0 the time-stamp channel
                if data.shape[0] < irf.shape[1]:
                    irf[cursor, :data.shape[0], :] = data[:, :-1]
                    irf[cursor, data.shape[0]:, :] = data[-1, :-1]  # held with last value
                else:
                    irf[cursor, :, :] = data[:irf.shape[1], :-1]
                #print("seglen={}".format(data.shape[0]))
                irf_lab[cursor, ...] = stimulus
                # move the cursor
                cursor = (cursor + 1) % irf.shape[0]
            else:
                new_pending.append(stimulus_segment)
        # update the pending segments list
        pending = new_pending

        # Update the plots for each event type
        for ei, lab in enumerate(evtlabs):

            # extract this events irfs
            idx = np.flatnonzero(irf_lab[:, ei])
            # skip if no events to show
            if idx.size == 0:
                continue
            data = irf[idx, ...].copy()  #(nevt,nsamp,d)
            
            if center:
                data = data - data[:, 0:1, :]  # np.mean(data,axis=-2,keepdims=True)

            # compute ERP            
            erp = np.mean(data, 0)  # (nsamp,d)
            # plot ERP
            erp_lim = (np.min(erp.ravel()),np.max(erp.ravel()))
            erp_lines[ei].set_data(erp.T)
            erp_lines[ei].set_clim(vmin=erp_lim[0],vmax=erp_lim[1])
            erp_ax[ei].set_title("{} (n={})".format(lab,data.shape[0]))

            #print("erp={}".format(erp))
            # decompose into spatial and temporal components
            if True:
                # TODO: use multiCCA to get the decomp?
                R, s, A = np.linalg.svd(erp, full_matrices=False)
                A = A.T # (d,rank)
                # get the biggest eigenvalue to display
                slmidx = np.argsort(s) # N.B. ascending order
                slmidx = slmidx[::-1]  # N.B. DESCENDING order
                R = R[:, slmidx[:rank]]  #(nsamp,rank)
                s = s[slmidx[:rank]]
                A = A[:,slmidx[:rank]] * s[np.newaxis,:]  #(d,rank)
            else:
                A = np.zeros((erp.shape[-1])) 
                A[-1] = 1  #(1,d)
                R = erp[:, -1:]  # @ A
                
            # apply the spatial filter to the raw data to get pure time-course
            for ri in range(A.shape[-1]):
                # plot the spatial pattern
                spatial_lines[ei][ri][-1].set_ydata(A[:,ri])
                spatial_ax[ei][ri].set_ylim((np.min(A), np.max(A)))
                #temporal_ax[ei].plot(irf_times, R, color=(0,0,0), label='erp ({})'.format(data.shape[0]))

                # plot single-trial data
                #datari = data @ A[:,ri] # (d,1)?
                #for di in range(min(len(temporal_lines[ei][ri])-1, data.shape[0])):
                #    temporal_lines[ei][ri][-1-di-1].set_ydata(datari[-1-di, :])
                    
                # plot the average response, as thick black                
                temporal_lines[ei][ri].set_ydata(R[:,ri])
                if ri==0:
                    temporal_ax[ei].set_title("{} ({})".format(lab, data.shape[0]))
                    temporal_ax[ei].set_ylim( (np.min(R.ravel())*2, np.max(R.ravel()) * 2) )

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, help='address (IP) of the utopia-hub', default=None)
    parser.add_argument('--evtlabs', type=str, help='comma separated list of stimulus even types to use', default='re,fe')
    parser.add_argument('--rank', type=str, help='rank of decomposition to use', default=3)
    parser.add_argument('--ch_names', type=str, help='list of channel names, or capfile', default=None)
    args = parser.parse_args()
    hostname = args.host
    evtlabs = args.evtlabs.split(',')
    rank = args.rank
    print('evtlabs={}'.format(evtlabs))
    ch_names = args.ch_names.split(',') if args.ch_names is not None else None

    data_preprocessor = None
    #data_preprocessor = butterfilt_and_downsample(order=6, stopband='butter_stopband((0, 5), (25, -1))_fs200.pk', fs_out=60)
    data_preprocessor = butterfilt_and_downsample(order=4, stopband=((0, 4), (25, -1)), fs_out=80)
    ui=UtopiaDataInterface(data_preprocessor=data_preprocessor, send_signalquality=False)
    ui.connect(hostname)

    try:
        matplotlib.use('Qt4Cairo')
    except:
        pass

    erpViewer(ui, evtlabs=evtlabs, ch_names=ch_names, rank=rank)



from mindaffectBCI.examples.acquisation.cyton import OpenBCICyton
import asyncore # needed for the openBCI server
import numpy as np
from time import time, sleep
import socket

# add utopiaclient to path and import
import sys, os
from mindaffectBCI import utopiaclient

LOGINTERVAL_S = 2
SCALE_FACTOR_EEG = 1000000 * 4.5 / 24 / (2**23 - 1)
SCALE_FACTOR_AUX = .002 / 2**4

lastdata = None # last valid data seen
lastsamp = None # last sample id processed
lastcount = None # number of times we have seen this sample id -- for multi-sample packets
t0 = None # start time
nextLogTime = None # time print next log message
nSamp = 0 # number samples processed so far
ftBlockSample = 200/20 # minimum number samples in each block sent, so stream @20hz
nBlock = 0 # number blocks sent to ft-buffer
ftBlock = [] # storage for the current block samples to go to ft-buffer
LOGFILE = None # "ganglion.log"
fSample = None
samp0   = None

def getTime_ms():
    if client:
        return client.getTimeStamp()
    return None
def getTime():
    return getTime_ms()/1000

def incsampCount(sample, count):
    ''' wrap arround sample incrementer, with repeate tracker '''
    sample = sample+1 if sample < 255 else 0
    return (sample, count)

nerr = 0 # number of samples with sample underestimate
estndropped = 0 # estimated magnitude of the missing samples
epsilon_err = 1 # size of missing sample deadzone
thresh_nerr_s = 6 # number seconds consistently to few samples to padd
thresh_track_s = [6, 300] # time-constants for tracking slow drift in offset
def trackDroppedSamples(nsamp, ts_ms):
    '''function to detect dropped samples so can insert padding'''
    global nerr, estndropped, samp0
    if nsamp < fSample*10: # warm-up the offset
        samp0est = nsamp - ts_ms*fSample/1000
        samp0 = max(samp0, samp0est) if not samp0 is None else samp0est
        estndropped = fSample*10
        return 0

    # update the tracking info
    samp_est = int(ts_ms/1000*fSample + samp0)
    err = samp_est - nsamp
    # slow tracking of sample offset
    if err < 0: # proportional step for neg err => quadratic-lowerbound
        samp0 = samp0 - err/(thresh_track_s[0]*fSample)
    else: # fixed step for positive err => linear upper-bound
        samp0 = samp0 - min(1, err)/(thresh_track_s[1]*fSample)

    if err < epsilon_err:
        # inside deadzone
        nerr = 0
        return 0

    # outside the deadzone, i.e. err>=epsilon_err
    # track the total time with consistently fewer samples than expected
    nerr = nerr + 1
    
    if nerr > thresh_nerr_s*fSample/2: # record the size of the offset
        estndropped = min(estndropped, err)

    # if long enough with missing samples -> mark to insert the missing
    if nerr > thresh_nerr_s*fSample:
        nerr = 0
        ndropped = estndropped
        estndropped = fSample*10 # start with inf
        return ndropped # return the delta to the error
    return 0

def processandpadsample(sample):
    global nSamp, ftBlock, lastdata, lastsamp, lastcount
    data = [d*SCALE_FACTOR_EEG for d in sample.channels_data] # [ nEEG ]
    sample.sample_number = sample.id

    # get the sample number we are looking for
    if lastsamp:
      nextsamp, nextcount = incsampCount(lastsamp, lastcount)
    else:
      nextsamp = sample.sample_number
      nextcount = 0 # always start with 1st sample in packet

    if len(data) == 0:
        print("%d) Packet with no channel data, skipped!"%(nSamp))
        return
    
    # check for missing samples, and pad if necessary   
    if not sample.sample_number == nextsamp:
        print("%d) skipped a sample! (%d, %d)->%d, padding: "%(nSamp, lastsamp, lastcount, sample.sample_number), end = '')
        if not lastdata is None:
            while not nextsamp == sample.sample_number:
                ftBlockAppend(lastdata, nextsamp, nextcount, 's')
                nextsamp, nextcount = incsampCount(nextsamp, nextcount)
                print("s", end="")
            print()

    # second check based on tracking the sample arrival time
    ndroppedsamp = trackDroppedSamples(nSamp, getTime_ms())
    if ndroppedsamp:
        print("%d) %d dropped samples detected, padding: "%(nSamp, ndroppedsamp))
        if not lastdata is None:
            for _ in range(ndroppedsamp):
                ftBlockAppend(lastdata, lastsamp, lastcount, 'd')
                print("d", end="")
            print()
        print("%d %d: (%d,%d)->%d end-dropped"%(getTime_ms(), nSamp, lastsamp, lastcount, sample.sample_number))
      
    # finally put the actual new data
    # put this data sample and updated tracking info
    ftBlockAppend(data, sample.sample_number, None)


def ftBlockAppend(data, sample_number, sample_count=None, why=0):
    '''function to add a new data sample to the block data to send to ftbuffer,
        and log info about where the data came from'''
    global nSamp, ftBlock, lastdata, lastsamp, lastcount
    ftBlock.append(data)
    nSamp = nSamp + 1
    # store info on the last data we have processed
    lastdata = data
    if sample_count is None:
        lastcount = 1 if sample_number == lastsamp else 0
    else:
        lastcount = sample_count
    lastsamp = sample_number
    if LOGFILE:
        why = ord(why) if isinstance(why, str) else why
        print("%d %d %d %d %d"%(getTime_ms(), nSamp, why, lastsamp, lastcount), file=LOGFILE)
    return nSamp
    
def utopia_putsamples(samples):
    '''process multiple samples at a time'''
    global nSamp, nBlock, ftBlock
    if hasattr(samples, '__iter__'): # set samples
        for sample in samples:
            processandpadsample(sample)
    else: # single sample
        processandpadsample(samples)

    if len(ftBlock) > ftBlockSample:
        client.getNewMessages(0) # flush the incomming message queue
        client.sendMessage(utopiaclient.DataPacket(getTime_ms(), ftBlock))
        nBlock = nBlock+1
        ftBlock = []
    printLog(nSamp, nBlock)

def printLog(nSamp, nBlock):
    ''' textual logging of the data arrivals etc.'''
    global t0, nextLogTime
    if getTime() > nextLogTime:
        elapsed = getTime()-t0
        print("%d %d %f %f (samp,blk,s,hz)"%(nSamp, nBlock, elapsed, nSamp/elapsed), flush=True)
        nextLogTime = getTime()+LOGINTERVAL_S

board = None
client = None
def initConnections(host=None, obciport=None):
    global board, client, fSample, nChans
    print("Opening OpenBCI Cyton on {}".format(obciport)) 
    # N.B. latency is min time between packets in us -> 1 packet = 5 samples -> 50hz
    #      high_speed => binary format
    #      sample_rate => requested sample rate
    board = OpenBCICyton(port=obciport)
    
    # Get the board info    
    nChans = 16 if board.daisy else 8
    fSample = 250
    print("Opened openBCI on %s with %d ch @ %g hz"%(board.port, nChans, fSample))

    client = utopiaclient.UtopiaClient()
    client.disableHeartbeats() # disabled for data sources
    client.autoconnect(host)
    # don't subscribe to anything
    client.sendMessage(utopiaclient.Subscribe(getTime_ms(), ""));
    print("Putting header.")    
    client.sendMessage(utopiaclient.DataHeader(getTime_ms(), nChans, fSample, ""))
    return (board, client)

def parse_arguments():
    import sys
    argv = sys.argv
    print("Arguments:",*argv)
    # N.B. argv0 is the script name
    host=argv[1] if len(argv)>1 else 'localhost'
    obciport=argv[2] if len(argv)>2 else 'com4'
    return dict(host=host, obciport=obciport)

def run(host='localhost', obciport='com4', **kwargs):
    initConnections(host=host, obciport=obciport)
    # record the stream start time
    global t0, nextLogTime
    t0 = getTime()
    nextLogTime = t0
    
    global LOGFILE
    if LOGFILE:
        LOGFILE = open(LOGFILE, "w")
    # start the forwarder
    print("Starting the data forwarder")
    board.start_stream(utopia_putsamples)

if __name__ == "__main__":
    args=parse_arguments()
    run(**args)



import argparse
from brainflow.board_shim import BoardShim, BrainFlowInputParams
import FieldTrip
from time import time, sleep
import traceback

PACKETRATE_HZ = 50
LOGINTERVAL_S = 3
t0=None
nextLogTime=None
def printLog(nSamp, nBlock):
    ''' textual logging of the data arrivals etc.'''
    global t0, nextLogTime
    t = time()
    if t0 is None: 
        t0 = t
        nextLogTime = t
    if t > nextLogTime:
        elapsed = time()-t0
        print("%d %d %f %f (samp,blk,s,hz)"%(nSamp, nBlock, elapsed, nSamp/elapsed), flush=True)
        nextLogTime = t +LOGINTERVAL_S


def parse_args():
    parser = argparse.ArgumentParser ()
    # use docs to check which parameters are required for specific board, e.g. for Cyton - set serial port
    parser.add_argument ('--host', type = str, help  = 'host name for the utopia hub', required = False, default = None)
    parser.add_argument ('--timeout', type = int, help  = 'timeout for device discovery or connection', required = False, default = 0)
    parser.add_argument ('--ip-port', type = int, help  = 'ip port', required = False, default = 0)
    parser.add_argument ('--ip-protocol', type = int, help  = 'ip protocol, check IpProtocolType enum', required = False, default = 0)
    parser.add_argument ('--ip-address', type = str, help  = 'ip address', required = False, default = '')
    parser.add_argument ('--serial-port', type = str, help  = 'serial port', required = False, default = 'com3') # '')
    parser.add_argument ('--mac-address', type = str, help  = 'mac address', required = False, default = '')
    parser.add_argument ('--other-info', type = str, help  = 'other info', required = False, default = '')
    parser.add_argument ('--streamer-params', type = str, help  = 'streamer params', required = False, default = '')
    parser.add_argument ('--serial-number', type = str, help  = 'serial number', required = False, default = '')
    parser.add_argument ('--board-id', type = int, help  = 'board id, check docs to get a list of supported boards', default = 1 )#required = True)
    parser.add_argument ('--log', type=int, help = ' set the brainflow logging level', default=1)
    args = parser.parse_args ()
    return args

board = None
client = None
def run (host=None,board_id=1,ip_port=0,serial_port='',mac_address='',other_info='',
         serial_number='',ip_address='',ip_protocol=0,timeout=0,streamer_params='',log=0):
    global board, client

    # init the board params
    params = BrainFlowInputParams ()
    params.serial_port = serial_port
    params.ip_port = ip_port
    params.mac_address = mac_address
    params.other_info = other_info
    params.serial_number = serial_number
    params.ip_address = ip_address
    params.ip_protocol = ip_protocol
    params.timeout = timeout

    print('board_id={} params= {}'.format(board_id, vars(params)))

    if (log):
        BoardShim.enable_dev_board_logger ()
    else:
        BoardShim.disable_board_logger ()

    board = BoardShim (board_id , params)
    board.prepare_session ()

    eeg_channels = BoardShim.get_eeg_channels (board_id)
    timestamp_channel = BoardShim.get_timestamp_channel(board_id)
    fSample = BoardShim.get_sampling_rate(board_id)

    print("board with {} ch @ {} hz".format(len(eeg_channels), fSample))

    # connect to the utopia client
    if host is None: host='localhost' 
    port=1972
    ftc=FieldTrip.Client()
    while not ftc.isConnected:
        try:
            ftc.connect(host, port)
        except ConnectionRefusedError:
            print("Failed to connect at %s:%d"%(host,port))
            sleep(1)
    print("Connected to buffer: %s:%d"%(buffhost,port))
    print("Putting header.")    
    ftc.putHeader(len(eeg_channels), fSample, 9)

    sleep(1) # just in case
    board.start_stream (45000, streamer_params)
    # N.B. we force a sleep here to allow the board to startup correctly
    sleep(3)

    maxpacketsamples = int(32000 / 4 / len(eeg_channels))
    nSamp=0
    nBlock=0
    data=None
    ots = None
    while True:

        data = board.get_board_data () # (channels,samples) get all data and remove it from internal buffer
        if data.size == 0:
            sleep(.001)
            continue

        # BODGE: simulated boards seem to not work correctly without some blocking IO
        if board_id < 0:
            print('.',end='',flush=True)
        #print("{} samp in {} uS - {}".format(data.shape[1],dts,data.shape[1]/dts))

        # extract the info we want
        eeg = data[eeg_channels,:] # (channels,samples) 
        timestamps = data[timestamp_channel,:] # (samples,)

        # forward the *EEG* data to the utopia client
        nSamp = nSamp + eeg.shape[1]
        # format for sending to MA
        eeg = eeg.T # MA uses (samples,channels)

        # TODO[]: send as smaller packets if too much data
        if eeg.shape[0] < maxpacketsamples:
            # fit time-stamp into 32-bit int (with potential wrap-around)
            ts = timestamps[-1]
            ts = (int(ts*1000))%(1<<31) 
            #ts = client.getTimeStamp()
            ftc.putData(eeg)
            nBlock = nBlock + 1
        else:
            pktidx = list(range(0,eeg.shape[0],maxpacketsamples)) + [eeg.shape[0],]
            for i in range(len(pktidx)-1):
                d = eeg[pktidx[i]:pktidx[i+1],:]
                ts = timestamps[pktidx[i+1]-1]
                # ensure increasing time-stamps.... (brainflow bug?)
                ts = max(ots,ts) if ots is not None else ts
                ots = ts
                # fit time-stamp into 32-bit int (with potential wrap-around)
                ts = (int(ts*1000))%(1<<31) 
                #ts = client.getTimeStamp()
                ftc.putData(d)
                nBlock = nBlock + 1

        # limit the packet sending rate..
        sleep(1/PACKETRATE_HZ)
        printLog(nSamp,nBlock)        

    board.stop_stream ()
    board.release_session ()

if __name__ == "__main__":
    #run(board_id=1,serial_port='com3') # ganglion
    run(board_id=0,serial_port='com4')

    args=parse_args()    
    try:
        run(**vars(args))
    except:
        traceback.print_exc()



import numpy as np
from scipy.io import loadmat
from mindaffectBCI.decoder.utils import butter_sosfilt, window_axis, block_randomize
from mindaffectBCI.decoder.multipleCCA import robust_whitener
from mindaffectBCI.decoder.updateSummaryStatistics import updateCxx
import matplotlib.pyplot as plt

def load_ninapro_db2(datadir, stopband=((0,15), (45,65), (95,125), (250,-1)), envelopeband=(10,-1), trlen_ms=None, fs_out=60, nvirt=20, rectify=True, whiten=True, log=True, plot=False, filterbank=None, zscore_y=True, verb=1):
    d = loadmat(datadir, variable_names=('emg', 'glove', 'stimulus'))
    X = d['emg'] # (nSamp,d)
    Y = d['glove'] # (nSamp,e)
    lab = d['stimulus'].ravel() # (nSamp,1) - use to slice out trials+labels
    fs = 2000
    if fs_out is  None:
        fs_out = fs

    # get trial start/end info
    trl_start = np.flatnonzero(np.diff(lab)>0)
    lab = lab[trl_start+1]
    print('trl_start={}'.format(trl_start))
    print('label={}'.format(lab))
    print("diff(trl_start)={}".format(np.diff(trl_start)))
    if trlen_ms is None:
        trlen_ms = np.max(np.diff(trl_start))*1000/fs
        print('trlen_ms={}'.format(trlen_ms))

    if not stopband is None:
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs)
        if plot:plt.figure(101);plt.plot(X);plt.title("hp+notch+lp")
        # preprocess -> spatial whiten
        # TODO[] : make this  fit->transform method

    if whiten:
        if verb > 0:print("spatial whitener")
        Cxx = updateCxx(None,X,None)
        W,_ = robust_whitener(Cxx)
        X = np.einsum("sd,dw->sw",X,W)
        if plot:plt.figure(102);plt.plot(X);plt.title("+whiten")            

    if not filterbank is None:
        if verb > 0:  print("Filterbank: {}".format(filterbank))
        # apply filter bank to frequency ranges into virtual channels
        Xs=[]
        # TODO: make a nicer shape, e.g. (tr,samp,band,ch)
        for bi,band in enumerate(filterbank):
            Xf, _, _ = butter_sosfilt(X,band,fs)
            Xs.append(Xf)
        # stack the bands as virtual channels
        X = np.concatenate(Xs,-1)
        
    X = np.abs(X) # rectify

    if log:
        if verb > 0: print("log amplitude")
        X = np.log(np.maximum(X,1e-6))
    if plot:plt.figure(103);plt.plot(X);plt.title("+abs")
    if envelopeband is not None:
        if verb>0: print("Envelop band={}".format(envelopeband))
        X, _, _ = butter_sosfilt(X,envelopeband,fs) # low-pass = envelope extraction
        if plot:plt.figure(104);plt.plot(X);plt.title("env")

    # preprocess -> downsample
    resamprate = int(fs/fs_out)
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs, fs/resamprate, resamprate))
        X = X[..., ::resamprate, :] # decimate X (trl, samp, d)
        Y = Y[..., ::resamprate, :] # decimate Y (trl, samp, e)
        trl_start = trl_start/resamprate
        fs = fs/resamprate

    # pre-process : z-trans Y
    if zscore_y:
        if verb > 0:print("Z-trans Y")
        mu = np.mean(Y, axis=-2, keepdims=True)
        std = np.std(Y, axis=-2, keepdims=True)
        std[std < 1e-6] = 1 # guard divide by 0
        Y = (Y - mu) / std

    # generate artificial other stimulus streams, for testing
    # TODO: randomize in better way
    Y = Y[:, np.newaxis, :] # (nSamp, nY, e)
    Y_test = block_randomize(Y, nvirt, axis=-3, block_size=Y.shape[0]//100//2)
    Y = np.concatenate((Y, Y_test), -2) # (nSamp, nY, e)
   
    # slice X,Y into trials
    oX = X # (nSamp,d)
    oY = Y # (nSamp,nY,e)
    trlen_samp = int(trlen_ms * fs / 1000)
    X = np.zeros((trl_start.size, trlen_samp, X.shape[-1]))
    Y = np.zeros((trl_start.size, trlen_samp) + Y.shape[-2:])
    print("Slicing {} trials of {}ms".format(len(trl_start),trlen_ms))
    for ti,tii in enumerate(trl_start):
        tii = int(tii)
        trl_len = min(oX.shape[0],tii+trlen_samp) - tii
        X[ti, :trl_len, ...] = oX[tii:tii+trl_len, ...]
        Y[ti, :trl_len, ...] = oY[tii:tii+trl_len, ...]

    # make meta-info
    coords = [None]*X.ndim
    coords[0] = {'name':'trial', 'coords':lab}
    coords[1] = {'name':'time', 'fs':fs, 'units':'ms', 'coords':np.arange(X.shape[1])*1000/fs}
    coords[2] = {'name':'channel', 'coords':None}

    return (X, Y, coords)

def testcase():
    try:
        from offline.load_ninapro_db2 import load_ninapro_db2
    except:
        pass
    
    datadir = '/home/jadref/data/bci/external_data/ninapro/DB2_s1/S1_E1_A1.mat'

    fs_out = 30
    X, Y, coords = load_ninapro_db2(datadir, fs_out=fs_out, envelopeband=((0,.01),(5,-1)), log=True)
    from analyse_datasets import debug_test_dataset
    debug_test_dataset(X,Y,coords, rank=5, evtlabs=None, tau_ms=120, offset_ms=0, model='cca')
    
if __name__=="__main__":
    testcase()



from mindaffectBCI.decoder.utils import idOutliers, butter_sosfilt
from mindaffectBCI.decoder.updateSummaryStatistics import updateCxx
from mindaffectBCI.decoder.multipleCCA import robust_whitener
import numpy as np

def preprocess(X, Y, coords, whiten=False, whiten_spectrum=False, badChannelThresh=None, badTrialThresh=None, center=False, car=False, stopband=None, filterbank=None):
    """apply simple pre-processing to an input dataset

    Args:
        X ([type]): the EEG data (tr,samp,d)
        Y ([type]): the stimulus (tr,samp,e)
        coords ([type]): [description]
        whiten (float, optional): if >0 then strength of the spatially regularized whitener. Defaults to False.
        whiten_spectrum (float, optional): if >0 then strength of the spectrally regularized whitener. Defaults to False.
        badChannelThresh ([type], optional): threshold in standard deviations for detection and removal of bad channels. Defaults to None.
        badTrialThresh ([type], optional): threshold in standard deviations for detection and removal of bad trials. Defaults to None.
        center (bool, optional): flag if we should temporally center the data. Defaults to False.
        car (bool, optional): flag if we should spatially common-average-reference the data. Defaults to False.

    Returns:
        X ([type]): the EEG data (tr,samp,d)
        Y ([type]): the stimulus (tr,samp,e)
        coords ([type]): meta-info for the data
    """    
    if center:
        X = X - np.mean(X, axis=-2, keepdims=True)

    if badChannelThresh is not None:
        X, Y, coords = rmBadChannels(X, Y, coords, badChannelThresh)

    if badTrialThresh is not None:
        X, Y, coords = rmBadTrial(X, Y, coords, badTrialThresh)

    if car:
        print('CAR')
        X = X - np.mean(X, axis=-1, keepdims=True)

    if whiten>0:
        reg = whiten if not isinstance(whiten,bool) else 0
        print("whiten:{}".format(reg))
        X, W = spatially_whiten(X,reg=reg)

    if stopband is not None:
        X, _, _ = butter_sosfilt(X,stopband,fs=coords[-2]['fs'])

    if whiten_spectrum > 0:
        reg = whiten_spectrum if not isinstance(whiten_spectrum,bool) else 0
        print("Spectral whiten:{}".format(reg))
        X, W = spectrally_whiten(X, axis=-2, reg=reg)

    if filterbank is not None:
        X, _, _ = butter_filterbank(X,filterbank,fs=coords[-2]['fs'])
        # make filterbank entries into virtual channels
        X = np.reshape(X,X.shape[:-2]+(prod(X.shape[-2:],)))
        # update meta-info
        if 'coords' in coords[-1] and coords[-1]['coords'] is not None:
            ch_names = coords[-1]['coords']
            ch_names = ["{}_{}".format(c,f) for f in filterbank for c in coords]

    return X, Y, coords


def rmBadChannels(X:np.ndarray, Y:np.ndarray, coords, thresh=3.5):
    """remove bad channels from the input dataset

    Args:
        X ([np.ndarray]): the eeg data as (trl,sample,channel)
        Y ([np.ndarray]): the associated stimulus info as (trl,sample,stim)
        coords ([type]): the meta-info about the channel
        thresh (float, optional): threshold in standard-deviations for removal. Defaults to 3.5.

    Returns:
        X (np.ndarray)
        Y (np.ndarray)
        coords
    """    
    isbad, pow = idOutliers(X, thresh=thresh, axis=(0,1))
    print("Ch-power={}".format(pow.ravel()))
    keep = isbad[0,0,...]==False
    X=X[...,keep]

    if 'coords' in coords[-1] and coords[-1]['coords'] is not None:
        rmd = coords[-1]['coords'][isbad[0,0,...]]
        print("Bad Channels Removed: {} = {}={}".format(np.sum(isbad),np.flatnonzero(isbad[0,0,...]),rmd))

        coords[-1]['coords']=coords[-1]['coords'][keep]
    else:
        print("Bad Channels Removed: {} = {}".format(np.sum(isbad),np.flatnonzero(isbad[0,0,...])))

    if 'pos2d' in coords[-1] and coords[-1]['pos2d'] is not None:  
        coords[-1]['pos2d'] = coords[-1]['pos2d'][keep]

    return X,Y,coords

def rmBadTrial(X, Y, coords, thresh=3.5, verb=1):
    """[summary]

    Args:
        X ([np.ndarray]): the eeg data as (trl,sample,channel)
        Y ([np.ndarray]): the associated stimulus info as (trl,sample,stim)
        coords ([type]): the meta-info about the channel
        thresh (float, optional): threshold in standard-deviations for removal. Defaults to 3.5.
    Returns:
        X (np.ndarray)
        Y (np.ndarray)
        coords
    """
    isbad,pow = idOutliers(X, thresh=thresh, axis=(1,2))
    print("Trl-power={}".format(pow.ravel()))
    X=X[isbad[...,0,0]==False,...]
    Y=Y[isbad[...,0,0]==False,...]

    if 'coords' in coords[0] and np.sum(isbad) > 0:
        rmd = coords[0]['coords'][isbad[...,0,0]]
        print("BadTrials Removed: {} = {}".format(np.sum(isbad),rmd))

        coords[0]['coords']=coords[0]['coords'][isbad[...,0,0]==False]
    return X,Y,coords


def spatially_whiten(X:np.ndarray, *args, **kwargs):
    """spatially whiten the nd-array X

    Args:
        X (np.ndarray): the data to be whitened, with channels/space in the *last* axis

    Returns:
        X (np.ndarray): the whitened X
        W (np.ndarray): the whitening matrix used to whiten X
    """    
    Cxx = updateCxx(None,X,None)
    W,_ = robust_whitener(Cxx, *args, **kwargs)
    X = X @ W #np.einsum("...d,dw->...w",X,W)
    return (X,W)

def spectrally_whiten(X:np.ndarray, reg=.01, axis=-2):
    """spatially whiten the nd-array X

    Args:
        X (np.ndarray): the data to be whitened, with channels/space in the *last* axis

    Returns:
        X (np.ndarray): the whitened X
        W (np.ndarray): the whitening matrix used to whiten X
    """    
    from scipy.fft import fft, ifft

    # TODO[]: add hanning window to reduce spectral leakage

    Fx = fft(X,axis=axis)
    H=np.abs(Fx)
    if Fx.ndim+axis > 0:
        H = np.mean(H, axis=tuple(range(Fx.ndim+axis))) # grand average spectrum (freq,d)

    # compute *regularized* whitener (so don't amplify low power noise components)
    W = 1./(H+np.max(H)*reg)

    # apply the whitener
    Fx = Fx * W 

    # map back to time-domain
    X = np.real(ifft(Fx,axis=axis))
    return (X,W)


def butter_filterbank(X:np.ndarray, filterbank, fs:float, axis=-2, order=4, ftype='butter', verb=1):
    if verb > 0:  print("Filterbank: {}".format(filterbank))
    if not axis == -2:
        raise ValueError("axis other than -2 not supported yet!")
    # apply filter bank to frequency ranges into virtual channels
    Xf=np.zeros(X.shape[:axis+1]+(len(filterbank),)+X.shape[axis+1:],dtype=X.dtype)
    for bi,stopband in enumerate(filterbank):
        if verb>1: print("{}) band={}\n".format(bi,stopband))
        Xf[...,bi,:], _, _ = butter_sosfilt(X.copy(),stopband=stopband,fs=fs,order=order,ftype=ftype)
    # TODO[X]: make a nicer shape, e.g. (tr,samp,band,ch)
    #X = np.concatenate([X[...,np.newaxis,:] for X in Xs],-2)
    return Xf


def plot_grand_average_spectrum(X, fs:float, axis:int=-2, ch_names=None, log=False):
    import matplotlib.pyplot as plt
    from scipy.signal import welch
    from mindaffectBCI.decoder.updateSummaryStatistics import plot_erp
    freqs, FX = welch(X, axis=axis, fs=fs, nperseg=fs//2, return_onesided=True, detrend=False) # FX = (nFreq, nch)
    print('FX={}'.format(FX.shape))
    #plt.figure(18);plt.clf()
    muFX = np.median(FX,axis=0,keepdims=True)
    if log:
        muFX = 10*np.log10(muFX)
        unit='db (10*log10(uV^2))'
        ylim = (2*np.median(np.min(muFX,axis=tuple(range(muFX.ndim-1))),axis=-1),
                2*np.median(np.max(muFX,axis=tuple(range(muFX.ndim-1))),axis=-1))
    else:
        unit='uV^2'
        ylim = (0,2*np.median(np.max(muFX,axis=tuple(range(muFX.ndim-1))),axis=-1))

    plot_erp(muFX, ch_names=ch_names, evtlabs=None, times=freqs, ylim=ylim)       
    plt.suptitle("Grand average spectrum ({})".format(unit))

def extract_envelope(X,fs,
                     stopband=None,whiten=True,filterbank=None,log=True,env_stopband=(10,-1),
                     verb=False, plot=False):
    """extract the envelope from the input data

    Args:
        X ([type]): [description]
        fs ([type]): [description]
        stopband ([type], optional): pre-filter stop band. Defaults to None.
        whiten (bool, optional): flag if we spatially whiten before envelope extraction. Defaults to True.
        filterbank ([type], optional): set of filters to apply to extract the envelope for each filter output. Defaults to None.
        log (bool, optional): flag if we return raw power or log-power. Defaults to True.
        env_stopband (tuple, optional): post-filter on the extracted envelopes. Defaults to (10,-1).
        verb (bool, optional): verbosity level. Defaults to False.
        plot (bool, optional): flag if we plot the result of each preprocessing step. Defaults to False.

    Returns:
        X: the extracted envelopes
    """                     
    from multipleCCA import robust_whitener
    from updateSummaryStatistics import updateCxx
    from utils import butter_sosfilt
    
    if plot:
        import matplotlib.pyplot as plt
        plt.figure(100);plt.clf();plt.plot(X[:int(fs*10),:].copy());plt.title("raw")

    if not stopband is None:
        if verb > 0:  print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs)
        if plot:plt.figure(101);plt.clf();plt.plot(X[:int(fs*10),:].copy());plt.title("hp+notch+lp")
        # preprocess -> spatial whiten
        # TODO[] : make this  fit->transform method
    
    if whiten:
        if verb > 0:  print("spatial whitener")
        Cxx = updateCxx(None,X,None)
        W,_ = robust_whitener(Cxx)
        X = np.einsum("sd,dw->sw",X,W)
        if plot:plt.figure(102);plt.clf();plt.plot(X[:int(fs*10),:].copy());plt.title("+whiten")            

    if not filterbank is None:
        X = filterbank(X,filterbank,fs)
        # make filterbank entries into virtual channels
        X = np.reshape(X,X.shape[:-2]+(prod(X.shape[-2:],)))
        
    X = np.abs(X) # rectify
    if plot:plt.figure(104);plt.plot(X[:int(fs*10),:]);plt.title("+abs")

    if log:
        if verb > 0: print("log amplitude")
        X = np.log(np.maximum(X,1e-6))
        if plot:plt.figure(105);plt.clf();plt.plot(X[:int(fs*10),:]);plt.title("+log")

    if env_stopband is not None:
        if verb>0: print("Envelop band={}".format(env_stopband))
        X, _, _ = butter_sosfilt(X,env_stopband,fs) # low-pass = envelope extraction
        if plot:plt.figure(104);plt.clf();plt.plot(X[:int(fs*10),:]);plt.title("+env")
    return X


def testCases():
    import numpy as np
    import matplotlib.pyplot as plt
    from mindaffectBCI.decoder.updateSummaryStatistics import plot_erp
    fs=100
    X = np.random.standard_normal((2,fs*3,2)) # flat spectrum
    X = X[:,:-1,:]+X[:,1:,:] # weak low-pass
    #X = np.cumsum(X,-2) # 1/f spectrum
    print("X={}".format(X.shape))
    #plt.figure()
    #plot_grand_average_spectrum(X, fs)
    #plt.show()

    bands = ((0,10,'bandpass'),(10,20,'bandpass'),(20,-1,'bandpass'))
    Xf = butter_filterbank(X,bands,fs,order=8,ftype='bessel') # tr,samp,band,ch
    print("Xf={}".format(Xf.shape))
    # bands -> virtual channels
    # make filterbank entries into virtual channels
    plt.figure()
    Xf_s = np.sum(Xf,-2,keepdims=False)
    plot_grand_average_spectrum(np.concatenate((X[:,np.newaxis,...],Xf_s[:,np.newaxis,...],np.moveaxis(Xf,(0,1,2,3),(0,2,1,3))),1), fs)
    plt.legend(('X','Xf_s','X_bands'))
    plt.show()
    
    # compare raw vs summed filterbank
    plt.figure()
    plot_erp(np.concatenate((X[0:1,...],Xf_s[0:1,...],np.moveaxis(Xf[0,...],(0,1,2),(1,0,2))),0),
             evtlabs=['X','Xf_s']+['Xf_{}'.format(b) for b in bands])
    plt.suptitle('X, Xf_s')
    plt.show()
    

if __name__=="__main__":
    testCases()
